# -*- encoding:utf-8 -*-
"""
    横截面下因子评估方式--因子绩效
"""
import copy, functools, pdb
import numpy as np
from collections import namedtuple
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress

DALIY_PER_YEAR = 252
WEEKLY_PER_YEAR = 52
MONTHLY_PER_YEAR = 12
QUARTERLY_PER_YEAR = 4
YEARLY_PER_YEAR = 1

OLNY_LONG = 'long'
OLNY_SHORT = 'short'
BOTH_SIDE = 'both'

POSITIVE = 1
NEGATIVE = -1


class EvaluateTuple(
        namedtuple('EvaluateTuple',
                   ('returns_mean', 'returns_std', 'sharp', 'turnover',
                    'maxdd', 'returns_mdd', 'win_rate', 'ic', 'ir', 'fitness',
                    'category', 'freq', 'hold'))):

    __slots__ = ()

    def __repr__(self):
        return "\nreturns_mean:{}\nreturns_std:{}\nsharp:{}\nturnover:{}\n" \
                "maxdd:{}\nreturns_mdd:{}\nwin_rate:{}\nic:{}\nir:{}\nfitness:{}\n" \
                "category:{}\nfreq:{}\nhold:{}".format(self.returns_mean,self.returns_std,
                    self.sharp,self.turnover,self.maxdd,self.returns_mdd,self.win_rate,
                    self.ic,self.ir,self.fitness,self.category,self.freq,self.hold)


class MetricsTuple(
        namedtuple('EvaluateTuple', ('long_evaluate', 'short_evaluate',
                                     'both_evaluate', 'hold', 'freq'))):
    __slots__ = ()

    def __repr__(self):
        return "long_evaluate:{}\nshort_evaluate:{}\nboth_evaluate:{},hold:{},freq:{}".format(
            self.long_evaluate, self.short_evaluate, self.both_evaluate,
            self.hold, self.freq)


def valid_check(func):
    """检测度量的输入是否正常，非正常显示info，正常继续执行被装饰方法"""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.valid:
            return func(self, *args, **kwargs)
        else:
            kd_logger.info('metrics input is invalid or zero order gen!')

    return wrapper


class Metrics(object):

    @classmethod
    def general(cls,
                returns,
                factors,
                hold=1,
                freq=DALIY_PER_YEAR,
                show_log=True):
        metrics = cls(returns=returns,
                      factors=factors,
                      hold=hold,
                      freq=freq,
                      show_log=show_log)
        return metrics.fit_metrics()

    def __init__(self,
                 returns,
                 factors,
                 hold=1,
                 freq=DALIY_PER_YEAR,
                 show_log=True):
        # 验证输入的数据是否可度量，便于valid_check装饰器工作
        self.valid = False
        self.returns = returns
        self.factors = factors
        self.freq = freq
        self.hold = hold
        self.direction = None
        self.show_log = show_log
        if self.returns is not None and self.factors is not None:
            self.valid = True

    @valid_check
    def fit_metrics(self):
        with Progress(3, 0, label='factor fit metrics') as progress:
            long_weight, short_weight, both_weight = self.create_weight()
            if self.show_log: progress.show(a_progress=1)
            long_evaluate = self.evaluate(returns=self.returns,
                                          weight=long_weight,
                                          freq=self.freq,
                                          hold=self.hold,
                                          category=OLNY_LONG)
            if self.show_log: progress.show(a_progress=2)
            short_evaluate = self.evaluate(returns=self.returns,
                                           weight=short_weight,
                                           freq=self.freq,
                                           hold=self.hold,
                                           category=OLNY_SHORT)
            if self.show_log: progress.show(a_progress=3)
            both_evaluate = self.evaluate(returns=self.returns,
                                          weight=both_weight,
                                          freq=self.freq,
                                          hold=self.hold,
                                          category=BOTH_SIDE)

            self.show_log: progress.show(a_progress=4)
        return MetricsTuple(long_evaluate=long_evaluate,
                            short_evaluate=short_evaluate,
                            both_evaluate=both_evaluate,
                            freq=self.freq,
                            hold=self.hold)

    def create_weight(self):
        right_weight = copy.deepcopy(self.factors)
        right_weight[right_weight <= 0] = np.nan
        right_weight = right_weight.div(right_weight.sum(axis=1, min_count=1),
                                        axis='rows')
        right_weight = right_weight.rolling(self.hold,
                                            min_periods=1).sum() / self.hold

        left_weight = copy.deepcopy(self.factors)
        left_weight[left_weight >= 0] = np.nan
        left_weight = left_weight.div(left_weight.sum(axis=1, min_count=1),
                                      axis='rows')
        left_weight = left_weight.rolling(self.hold,
                                          min_periods=1).sum() / self.hold

        both_weight = right_weight.sub(left_weight, fill_value=0)
        ret_diff = (self.returns * right_weight).sum(
            axis=1).mean() - (self.returns * left_weight).sum(axis=1).mean()
        if ret_diff > 0:
            self.direction = POSITIVE
            long_weight = copy.deepcopy(right_weight)
            short_weight = copy.deepcopy(left_weight)
        else:
            self.direction = NEGATIVE
            long_weight = copy.deepcopy(left_weight)
            short_weight = copy.deepcopy(right_weight)
            both_weight = copy.deepcopy(-both_weight)

        return long_weight, short_weight, both_weight

    def evaluate(self,
                 returns,
                 weight,
                 hold=1,
                 freq=DALIY_PER_YEAR,
                 category=OLNY_LONG):
        rets_sum = (returns * weight).sum(axis=1)
        rets_mean = rets_sum.mean() * freq
        rets_std = rets_sum.std() * np.sqrt(freq)
        sharp = rets_mean / rets_std
        tv = abs(weight.sub(weight.shift(1), fill_value=0)).sum(
            axis=1, min_count=1).mean() * 0.5
        pnl = rets_sum.cumsum()
        maxdd = (pnl.expanding().max() - pnl).max()
        ret2mdd = rets_mean / maxdd
        win_rate = rets_sum[
            rets_sum > 0].count() / rets_sum[~rets_sum.isna()].count()
        ic_series = weight.corrwith(returns, axis=1, method='spearman')
        ic = ic_series.mean()
        ir = ic / ic_series.std()
        fitness = sharp * np.sqrt(abs(rets_mean) / tv)

        return EvaluateTuple(returns_mean=rets_mean,
                             returns_std=rets_std,
                             sharp=sharp,
                             turnover=tv,
                             maxdd=maxdd,
                             returns_mdd=ret2mdd,
                             win_rate=win_rate,
                             ic=ic,
                             ir=ir,
                             fitness=fitness,
                             category=category,
                             freq=freq,
                             hold=hold)



# -*- encoding:utf-8 -*-
"""
    并行封装模块，主要针对不同平台统一接口规范：

    windows 上使用joblib进行长时间的多任务，如超过10小时以上时，在任何最后有系统pop任务
    的错误，所以windows上使用ProcessPoolExecutor进行多任务，套上Parallel和delayed保持接口的
    通用性及规范统一
"""

import warnings, functools

g_cpu_cnt = 4  # fix me
if False:
    """
        对外版本不再使用joblib避免python2，python3，mac, windows等joblib最后的pop一直出不来的bug，
        有能力可自行打开，在mac下python3环境且cpu足够快建议打开
    """
    # MAC 直接import Parallel, delayed
    # noinspection PyUnresolvedReferences
    from ..ExtBu.joblib import Parallel, delayed
else:
    # windows需要使用ProcessPoolExecutor
    try:
        # noinspection PyCompatibility
        from concurrent.futures import ProcessPoolExecutor
    except ImportError:
        warnings.warn("not installed ProcessPoolExecutor")

    def delayed(function):
        """
        将function通过functools.wraps及delayed_function进行保留，但不执行
        :param function:
        :return:
        """

        def delayed_function(*args, **kwargs):
            """将function以及参数返回为tuple，tuple[0]为原始function"""
            return function, args, kwargs

        try:
            delayed_function = functools.wraps(function)(delayed_function)
        except AttributeError:
            raise TypeError('wraps fails on some callable objects')
        return delayed_function

    class Parallel(object):
        """封装ProcessPoolExecutor进行并行任务执行操作"""

        def __init__(self,
                     n_jobs=1,
                     backend='multiprocessing',
                     verbose=0,
                     pre_dispatch='2 * n_jobs',
                     batch_size='auto',
                     temp_folder=None,
                     max_nbytes='1M',
                     mmap_mode='r'):
            """
            :param n_jobs: 并行启动的进程数，任务数量
            :param backend: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            :param verbose: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            :param pre_dispatch: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            :param batch_size: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            :param temp_folder: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            :param max_nbytes: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            :param mmap_mode: 无意义，只是为了统一接口规范，与joblib.Parallel保持一样的参数
            """
            self.n_jobs = n_jobs

        def __call__(self, iterable):
            """为与joblib并行保持一致，内部使用ProcessPoolExecutor开始工作"""

            result = []

            def when_done(r):
                """ProcessPoolExecutor每一个进程结束后结果append到result中"""
                result.append(r.result())

            if self.n_jobs <= 0:
                # 主要为了适配 n_jobs = -1，joblib中启动cpu个数个进程并行执行
                self.n_jobs = g_cpu_cnt

            if self.n_jobs == 1:
                # 如果只开一个进程，那么只在主进程(或当前运行的子进程)里运行，方便pdb debug且与joblib运行方式保持一致
                for jb in iterable:
                    result.append(jb[0](*jb[1], **jb[2]))
            else:
                with ProcessPoolExecutor(max_workers=self.n_jobs) as pool:
                    for jb in iterable:
                        # 这里iterable里每一个元素是delayed.delayed_function保留的tuple
                        future_result = pool.submit(jb[0], *jb[1], **jb[2])
                        future_result.add_done_callback(when_done)
            return result


def run_in_thread(func, *args, **kwargs):
    """
    多线程工具函数，不涉及返回值等细节处理时使用
    :param func: 被线程委托的函数
    :return: 返回Thread线程对象
    """
    from threading import Thread
    thread = Thread(target=func, args=args, kwargs=kwargs)
    thread.daemon = True
    thread.start()
    return thread


def run_in_subprocess(func, *args, **kwargs):
    """
    多进程工具函数，不涉及返回值等细节处理时使用
    :param func: 被进程委托的函数
    :return: 返回multiprocessing进程对象
    """
    from multiprocessing import Process
    process = Process(target=func, args=args, kwargs=kwargs)
    process.daemon = True
    process.start()
    return process


# -*- coding: utf-8 -*-
from typing import Optional
import numpy as np
import pandas as pd
from ultron.factor.utilities import agg_mean
from ultron.factor.data.quantile import quantile
from ultron.factor.data.standardize import standardize
from ultron.factor.data.winsorize import winsorize_normal
from ultron.factor.data.processing import factor_processing


def quantile_analysis(factors: pd.DataFrame,
                      factor_weights: np.ndarray,
                      dx_return: np.ndarray,
                      n_bins: int=5,
                      risk_exp: Optional[np.ndarray]=None,
                      **kwargs):

    if 'pre_process' in kwargs:
        pre_process = kwargs['pre_process']
        del kwargs['pre_process']
    else:
        pre_process = [winsorize_normal, standardize]

    if 'post_process' in kwargs:
        post_process = kwargs['post_process']
        del kwargs['post_process']
    else:
        post_process = [standardize]

    er = factor_processing(factors.values, pre_process, risk_exp, post_process) @ factor_weights
    return er_quantile_analysis(er, n_bins, dx_return, **kwargs)


def er_quantile_analysis(er: np.ndarray,
                         n_bins: int,
                         dx_return: np.ndarray,
                         de_trend=False) -> np.ndarray:

    er = er.flatten()
    q_groups = quantile(er, n_bins)

    if dx_return.ndim < 2:
        dx_return.shape = -1, 1

    group_return = agg_mean(q_groups, dx_return).flatten()
    total_return = group_return.sum()
    ret = group_return.copy()

    if de_trend:
        resid = n_bins - 1
        res_weight = 1. / resid
        for i, value in enumerate(ret):
            ret[i] = (1. + res_weight) * value - res_weight * total_return

    return ret


import functools
from collections import Iterable
import pandas as pd
from ultron.ump.core.fixes import six, partial

try:
    # noinspection PyUnresolvedReferences
    from pandas.core.window import EWM
    g_pandas_has_ewm = True
except ImportError:
    g_pandas_has_ewm = False

try:
    # noinspection PyUnresolvedReferences
    from pandas.core.window import Rolling
    g_pandas_has_rolling = True
except ImportError:
    g_pandas_has_rolling = False

try:
    # noinspection PyUnresolvedReferences
    from pandas.tseries.resample import DatetimeIndexResampler
    g_pandas_has_resampler = True
except ImportError:
    try:
        # noinspection PyUnresolvedReferences
        from pandas.core.resample import DatetimeIndexResampler
        g_pandas_has_resampler = True
    except ImportError:
        g_pandas_has_resampler = False

if pd.__version__ >= '0.20.0':
    g_pandas_has_ewm = True
    g_pandas_has_rolling = True


def __pd_object_covert_start(iter_obj):
    """
    _pd_object_covert中进行参数检测及转换
    :param iter_obj: 将要进行操作的可迭代序列
    :return: 操作之后的返回值是否需要转换为np.array
    """
    if isinstance(iter_obj, (pd.Series, pd.DataFrame)):
        # 如果本身就是(pd.Series, pd.DataFrame)，返回对返回值不需要转换，即False
        return iter_obj, False
    # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
    if isinstance(iter_obj,
                  Iterable) and not isinstance(iter_obj, six.string_types):
        # 可迭代对象使用pd.Series进行包装，且返回对返回值需要转换为np.array，即True
        return pd.Series(iter_obj), True
    raise TypeError('pd_object must support Iterable!!!')


def _pd_object_covert(func):
    """针对参数序列进行pandas处理的事前，事后装饰器"""

    @functools.wraps(func)
    def wrapper(pd_object, pd_object_cm, how, *args, **kwargs):
        """事前装饰工作__pd_object_covert_start，事后根据是否需要转换为np.array工作"""
        # 事前装饰工作__pd_object_covert_start
        pd_object, ret_covert = __pd_object_covert_start(pd_object)
        ret = func(pd_object, pd_object_cm, how, *args, **kwargs)
        # 事后根据是否需要转换为np.array工作
        if ret is not None and ret_covert:
            return ret.values
        return ret

    return wrapper


@_pd_object_covert
def _pd_rolling(pd_object, pd_object_cm, how, *args, **kwargs):
    """
    被_pd_object_covert装饰，对pandas中的rolling操作，根据pandas version版本自动选择调用方式
    :param pd_object: 可迭代的序列，pd.Series, pd.DataFrame或者只是Iterable
    :param pd_object_cm: 与pd_object相同，针对需要两个pandas对象或者序列执行的操作，如corr，cov等
    :param how: 代表方法操作名称，eg. mean, std, var
    :return:
    """
    if g_pandas_has_rolling:
        """pandas版本高，使用如pd_object.rolling直接调用"""
        rolling_obj = pd_object.rolling(*args, **kwargs)
        if hasattr(rolling_obj, how):
            if pd_object_cm is None:
                return getattr(rolling_obj, how)()
            # 需要两个pd_object进行的操作, getattr(rolling_obj, how)(pd_object_cm)
            return getattr(rolling_obj, how)(pd_object_cm)
    else:
        """pandas版本低，使用如pd.rolling_mean方法调用"""
        how_func = 'rolling_{}'.format(how)
        if hasattr(pd, how_func):
            if pd_object_cm is None:
                return getattr(pd, how_func)(pd_object, *args, **kwargs)
            # 需要两个pd_object进行的操作，getattr(pd, how_func)(pd_object, pd_object_cm, *args, **kwargs)
            return getattr(pd, how_func)(pd_object, pd_object_cm, *args,
                                         **kwargs)
    raise RuntimeError('_pd_rolling {} getattr error'.format(how))


"""没有全部导出，只导出常用的"""
pd_rolling_mean = partial(_pd_rolling, how='mean', pd_object_cm=None)
pd_rolling_median = partial(_pd_rolling, how='median', pd_object_cm=None)
pd_rolling_std = partial(_pd_rolling, how='std', pd_object_cm=None)
pd_rolling_var = partial(_pd_rolling, how='var', pd_object_cm=None)
pd_rolling_max = partial(_pd_rolling, how='max', pd_object_cm=None)
pd_rolling_min = partial(_pd_rolling, how='min', pd_object_cm=None)
pd_rolling_sum = partial(_pd_rolling, how='sum', pd_object_cm=None)
pd_rolling_kurt = partial(_pd_rolling, how='kurt', pd_object_cm=None)
pd_rolling_skew = partial(_pd_rolling, how='skew', pd_object_cm=None)
pd_rolling_corr = partial(_pd_rolling, how='corr')
pd_rolling_cov = partial(_pd_rolling, how='cov')


@_pd_object_covert
def _pd_ewm(pd_object, pd_object_cm, how, *args, **kwargs):
    """
    被_pd_object_covert装饰，对pandas中的ewm操作，根据pandas version版本自动选择调用方式
    :param pd_object: 可迭代的序列，pd.Series, pd.DataFrame或者只是Iterable
    :param pd_object_cm: 与pd_object相同，针对需要两个pandas对象或者序列执行的操作，如corr，cov等
    :param how: 代表方法操作名称，eg. mean, std, var
    :return:
    """
    if g_pandas_has_ewm:
        """pandas版本高，使用如pd_object.ewm直接调用"""
        ewm_obj = pd_object.ewm(*args, **kwargs)
        if hasattr(ewm_obj, how):
            if pd_object_cm is None:
                return getattr(ewm_obj, how)()
            # 需要两个pd_object进行的操作
            return getattr(ewm_obj, how)(pd_object_cm)
    else:
        """pandas版本低，使用如pd.ewmstd方法调用"""
        if how == 'mean':
            # pd.ewma特殊代表加权移动平均，所以使用a替换mean
            how = 'a'
        how_func = 'ewm{}'.format(how)
        if hasattr(pd, how_func):
            if pd_object_cm is None:
                return getattr(pd, how_func)(pd_object, *args, **kwargs)
            # 需要两个pd_object进行的操作
            return getattr(pd, how_func)(pd_object, pd_object_cm, *args,
                                         **kwargs)
    raise RuntimeError('_pd_ewm {} getattr error'.format(how))


"""没有全部导出，只导出常用的"""
pd_ewm_mean = partial(_pd_ewm, how='mean', pd_object_cm=None)
pd_ewm_std = partial(_pd_ewm, how='std', pd_object_cm=None)
pd_ewm_var = partial(_pd_ewm, how='var', pd_object_cm=None)
pd_ewm_corr = partial(_pd_ewm, how='corr')
pd_ewm_cov = partial(_pd_ewm, how='cov')


@_pd_object_covert
def _pd_rolling(pd_object, pd_object_cm, how, *args, **kwargs):
    """
    被_pd_object_covert装饰，对pandas中的rolling操作，根据pandas version版本自动选择调用方式
    :param pd_object: 可迭代的序列，pd.Series, pd.DataFrame或者只是Iterable
    :param pd_object_cm: 与pd_object相同，针对需要两个pandas对象或者序列执行的操作，如corr，cov等
    :param how: 代表方法操作名称，eg. mean, std, var
    :return:
    """
    if g_pandas_has_rolling:
        """pandas版本高，使用如pd_object.rolling直接调用"""
        rolling_obj = pd_object.rolling(*args, **kwargs)
        if hasattr(rolling_obj, how):
            if pd_object_cm is None:
                return getattr(rolling_obj, how)()
            # 需要两个pd_object进行的操作, getattr(rolling_obj, how)(pd_object_cm)
            return getattr(rolling_obj, how)(pd_object_cm)
    else:
        """pandas版本低，使用如pd.rolling_mean方法调用"""
        how_func = 'rolling_{}'.format(how)
        if hasattr(pd, how_func):
            if pd_object_cm is None:
                return getattr(pd, how_func)(pd_object, *args, **kwargs)
            # 需要两个pd_object进行的操作，getattr(pd, how_func)(pd_object, pd_object_cm, *args, **kwargs)
            return getattr(pd, how_func)(pd_object, pd_object_cm, *args,
                                         **kwargs)
    raise RuntimeError('_pd_rolling {} getattr error'.format(how))


"""没有全部导出，只导出常用的"""
pd_rolling_mean = partial(_pd_rolling, how='mean', pd_object_cm=None)
pd_rolling_median = partial(_pd_rolling, how='median', pd_object_cm=None)
pd_rolling_std = partial(_pd_rolling, how='std', pd_object_cm=None)
pd_rolling_var = partial(_pd_rolling, how='var', pd_object_cm=None)
pd_rolling_max = partial(_pd_rolling, how='max', pd_object_cm=None)
pd_rolling_min = partial(_pd_rolling, how='min', pd_object_cm=None)
pd_rolling_sum = partial(_pd_rolling, how='sum', pd_object_cm=None)
pd_rolling_kurt = partial(_pd_rolling, how='kurt', pd_object_cm=None)
pd_rolling_skew = partial(_pd_rolling, how='skew', pd_object_cm=None)
pd_rolling_corr = partial(_pd_rolling, how='corr')
pd_rolling_cov = partial(_pd_rolling, how='cov')


def pd_resample(pd_object, rule, *args, **kwargs):
    """
    对pandas中的resample操作，根据pandas version版本自动选择调用方式
    :param pd_object: 可迭代的序列，pd.Series, pd.DataFrame或者只是Iterable
    :param rule: 具体的resample函数中需要的参数 eg. 21D, 即重采样周期值
    :return:
    """
    if g_pandas_has_resampler:
        """pandas版本高，使用如pd_object.resample('21D').mean()直接调用"""
        how = kwargs.pop('how', '')
        rep_obj = pd_object.resample(rule)
        if hasattr(rep_obj, how):
            return getattr(rep_obj, how)()
        print('rep_obj how is error set!!!')
        return rep_obj
    else:
        """pandas版本低，使用如pd_object.resample('21D').how方法调用, 也就不用kwargs.pop('how', '')了"""
        return pd_object.resample(rule, *args, **kwargs)


# -*- encoding:utf-8 -*-
"""
    卖出择时策略因子基础模块
"""
from enum import Enum
from abc import ABCMeta, abstractmethod
from ultron.ump.core.fixes import six
from ultron.ump.core.base import ParamBase
from ultron.ump.slippage.sell_mean import SlippageSellMean
from ultron.ump.ump.manager import Manager
from ultron.ump.trade.ml_feature import MlFeature


class ESupportDirection(Enum):
    """子策略在support_direction中支持的方向数值定义"""
    DIRECTION_CAll = 1.0
    DIRECTION_PUT = -1.0


class FactorSellBase(six.with_metaclass(ABCMeta, ParamBase)):
    """
        卖出择时策略因子基类：卖出择时策略基类和买入择时基类不同，买入择时
        必须混入一个方向类，代表买涨还是买跌，且只能有一个方向，，卖出策略
        可以同时支持买涨，也可以只支持一个方向
    """

    def __init__(self, capital, kl_pd, combine_kl_pd, benchmark, **kwargs):
        """
        :param capital: 资金类AbuCapital实例化对象
        :param kl_pd: 择时时段金融时间序列，pd.DataFrame对象
        :param combine_kl_pd:合并了之前一年时间序列的金融时间序列，pd.DataFrame对象
        :param benchmark: 交易基准对象，AbuBenchmark实例对象, 因子可有选择性使用，比如大盘对比等功能
        """
        # 择时金融时间序列走势数据
        self.kl_pd = kl_pd
        # 机器学习特征数据构建需要，详情见make_sell_order_ml_feature中构造特征使用
        self.combine_kl_pd = combine_kl_pd
        # 资金情况数据
        self.capital = capital
        # 交易基准对象，AbuBenchmark实例对象, 因子可有选择性使用，比如大盘对比等功能
        self.benchmark = benchmark

        # 滑点类，默认AbuSlippageSellMean
        self.slippage_class = kwargs.pop('slippage', SlippageSellMean)

        # 构造ump对外的接口对象UmpManager
        self.ump_manger = Manager(self)

        # 默认的卖出说明，子类通过_init_self可覆盖更具体的名字
        self.sell_type_extra = '{}'.format(self.__class__.__name__)

        # 子类继续完成自有的构造
        self._init_self(**kwargs)

    def __str__(self):
        """打印对象显示：class name, slippage, kl_pd.info"""
        return '{}: slippage:{}, \nkl:\n{}'.format(self.__class__.__name__,
                                                   self.slippage_class,
                                                   self.kl_pd.info())

    __repr__ = __str__

    def read_fit_day(self, today, orders):
        """
        在择时worker对象中做日交易的函数，亦可以理解为盘前的一些决策事件处理，
        内部会调用子类实现的fit_day函数
        :param today: 当前驱动的交易日金融时间序列数据
         :param orders: 买入择时策略中生成的订单序列
        :return: 生成的交易订单AbuOrder对象
        """
        # 今天这个交易日在整个金融时间序列的序号
        self.today_ind = int(today.key)
        # 回测中默认忽略最后一个交易日
        if self.today_ind >= self.kl_pd.shape[0] - 1:
            return
        """
            择时卖出因子策略可支持正向，反向，或者两个方向都支持，
            针对order中买入的方向，filter策略,
            根据order支持的方向是否在当前策略支持范围来筛选order
        """
        orders = list(
            filter(
                lambda order: order.expect_direction in self.support_direction(
                ), orders))
        return self.fit_day(today, orders)

    def sell_tomorrow(self, order):
        """
        明天进行卖出操作，比如突破策略使用了今天收盘的价格做为参数，发出了买入信号，
        需要进行卖出操作，不能执行今天卖出操作
        :param order交易订单AbuOrder对象
        """
        order.fit_sell_order(self.today_ind, self)

    def sell_today(self, order):
        """
        今天即进行卖出操作，需要不能使用今天的收盘数据等做为fit_day中信号判断，
        适合如比特币非明确一天交易日时间或者特殊情况的卖出信号
        :param order交易订单AbuOrder对象
        """
        order.fit_sell_order(self.today_ind - 1, self)

    @abstractmethod
    def _init_self(self, **kwargs):
        """子类因子针对可扩展参数的初始化"""
        pass

    @abstractmethod
    def fit_day(self, today, orders):
        """子类主要需要实现的函数，完成策略因子针对每一个交易日的卖出交易策略"""
        pass

    def make_sell_order(self, order, day_ind):
        """
        根据交易发生的时间索引，依次进行：卖出交易时间序列特征生成，
        决策卖出交易是否拦截，生成特征学习数据，最终返回是否order成交，即订单生效
        :param order: 买入择时策略中生成的订单
        :param day_ind: 卖出交易发生的时间索引，即对应self.kl_pd.key
        :return:
        """

        # 卖出交易时间序列特征生成
        ml_feature_dict = self.make_sell_order_ml_feature(day_ind)
        # 决策卖出交易是否拦截
        block = self.make_ump_block_decision(ml_feature_dict)
        if block:
            return False

        # 如果卖出交易不被拦截，生成特征学习数据
        if order.ml_features is None:
            order.ml_features = ml_feature_dict
        else:
            order.ml_features.update(ml_feature_dict)
        return True

    # noinspection PyUnusedLocal
    def make_ump_block_decision(self, ml_feature_dict):
        """
        输入需要决策的当前卖出交易特征通过ump模块的对外manager对交易进行决策，
        判断是否拦截卖出交易，还是放行卖出交易。子类可复写此方法，即子类策略因子实现
        自己的任意ump组合拦截方式，根据策略的拦截比例需要等等参数确定ump具体策略，
        且对于多种策略并行执行策略本身定制适合自己的拦截策略，提高灵活度
        :param ml_feature_dict: 需要决策的当前卖出时刻交易特征dict
        :return:
        """
        return self.ump_manger.ump_block(ml_feature_dict)

    def make_sell_order_ml_feature(self, day_ind):
        """
         根据卖出交易发生的时间索引构通过AbuMlFeature构建卖出时刻的各个交易特征
         :param day_ind: 交易发生的时间索引，对应self.kl_pd.key
         :return:
         """
        return MlFeature().make_feature_dict(self.kl_pd,
                                             self.combine_kl_pd,
                                             day_ind,
                                             buy_feature=False)

    """TODO: 使用check support方式查询是否支持fit_week，fit_month，上层不再使用hasattr去判断"""
    # def fit_week(self, *args, **kwargs):
    #     pass

    # def fit_month(self, *args, **kwargs):
    #     pass


class FactorSellXD(FactorSellBase):
    """以周期为重要参数的策略，xd代表参数'多少天'如已周期为参数可直接继承使用 """

    def _init_self(self, **kwargs):
        """kwargs中必须包含: 突破参数xd 比如20，30，40天...突破"""
        # 向下突破参数 xd， 比如20，30，40天...突破
        self.xd = kwargs['xd']
        # 在输出生成的orders_pd中显示的名字
        self.sell_type_extra = '{}:{}'.format(self.__class__.__name__, self.xd)

    def read_fit_day(self, today, orders):
        """
        覆盖base函数, 为fit_day中切片周期金融时间序列数据
        :param today: 当前驱动的交易日金融时间序列数据
        :param orders: 买入择时策略中生成的订单序列
        :return: 生成的交易订单AbuOrder对象
        """
        # 今天这个交易日在整个金融时间序列的序号
        self.today_ind = int(today.key)

        # 回测中默认忽略最后一个交易日
        if self.today_ind >= self.kl_pd.shape[0] - 1:
            return
        orders = list(
            filter(
                lambda order: order.expect_direction in self.support_direction(
                ), orders))

        # 完成为fit_day中切片周期金融时间序列数据
        self.xd_kl = self.kl_pd[self.today_ind - self.xd + 1:self.today_ind +
                                1]
        return self.fit_day(today, orders)

    def support_direction(self):
        """raise NotImplementedError"""
        raise NotImplementedError('NotImplementedError support_direction')

    def fit_day(self, today, orders):
        """raise NotImplementedError"""
        raise NotImplementedError('NotImplementedError fit_day')



import os, copy, sys

os.environ['ULTRON_DATA'] = 'keim'
os.environ['IGNORE_WARNINGS'] = '0'
sys.path.insert(0, os.path.abspath('../'))
import pandas as pd
import numpy as np

from ultron.ump.technical.line import Line

from ultron.env import *

enable_example_env()

market_data = pd.read_csv('sz300059_20110728_20170726.csv', index_col=0)
market_data.head()

kl_tl = Line(market_data['close'][-252:], 'close')
rval = kl_tl.create_support_trend(only_last=True)


# -*- coding: utf-8 -*-

from ultron.tradingday.Enums.BizDayConventions import BizDayConventions
from ultron.tradingday.Enums.DateGeneration import DateGeneration
from ultron.tradingday.Enums.Factors import Factors
from ultron.tradingday.Enums.Months import Months
from ultron.tradingday.Enums.NormalizingType import NormalizingType
from ultron.tradingday.Enums.OptionType import OptionType
from ultron.tradingday.Enums.TimeUnits import TimeUnits
from ultron.tradingday.Enums.Weekdays import Weekdays

__all__ = ['BizDayConventions',
           'DateGeneration',
           'Factors',
           'Months',
           'NormalizingType',
           'OptionType',
           'TimeUnits',
           'Weekdays']



# -*- encoding:utf-8 -*-
"""
相对强弱指数（RSI）是通过比较一段时期内的平均收盘涨数和平均收盘跌数来分析市场买沽盘的意向和实力，
从而作出未来市场的走势

计算方法：

具体计算实现可阅读代码中_calc_rsi_from_pd()的实现
1. 根据收盘价格计算价格变动可以使用diff()也可以使用pct_change()
2. 分别筛选gain交易日的价格变动序列gain，和loss交易日的价格变动序列loss
3. 分别计算gain和loss的N日移动平均
4. rs = gain_mean / loss_mean
5. rsi = 100 - 100 / (1 + rs)

"""
import numpy as np
import pandas as pd

from ultron.ump.indicator.base import g_calc_type, ECalcType
from ultron.kdutils import scaler as scaler_util
from ultron.kdutils.decorator import catch_error
from ultron.ump.core.helper import pd_rolling_mean
"""_calc_rsi_from_pd计算rs时使用gain，否则使用change"""
g_rsi_gain = True


# noinspection PyUnresolvedReferences
def _calc_rsi_from_ta(prices, time_period=14):
    """
    使用talib计算rsi, 即透传talib.RSI计算结果
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: rsi的N日参数, 默认14
    """

    import talib
    if isinstance(prices, pd.Series):
        prices = prices.values
    rsi = talib.RSI(prices, timeperiod=time_period)
    return rsi


# noinspection PyTypeChecker
def _calc_rsi_from_pd(prices, time_period=14):
    """
    通过rsi公式手动计算rsi
    :param prices: 收盘价格序列，pd.Series或者np.array
    :param time_period: rsi的N日参数, 默认14
    """

    if not isinstance(prices, pd.Series):
        prices = pd.Series(prices)

    # 根据收盘价格计算价格变动可以使用diff()也可以使用pct_change()
    if g_rsi_gain:
        # 使用前后价格变动gain
        diff_price = prices.diff()
    else:
        # 使用前后价格变动change比例
        diff_price = prices.pct_change()
    diff_price[0] = 0

    # 分别筛选gain交易日的价格变动序列gain，和loss交易日的价格变动序列loss
    gain = np.where(diff_price > 0, diff_price, 0)
    loss = np.where(diff_price < 0, abs(diff_price), 0)
    # 分别计算gain和loss的N日移动平均
    gain_mean = pd_rolling_mean(gain, window=time_period)
    loss_mean = pd_rolling_mean(loss, window=time_period)
    # rsi = 100 - 100 / (1 +  gain_mean / loss_mean)
    rs = gain_mean / loss_mean
    rsi = 100 - 100 / (1 + rs)
    return rsi


"""通过在ABuNDBase中尝试import talib来统一确定指标计算方式, 外部计算只应该使用calc_rsi"""
calc_rsi = _calc_rsi_from_pd if g_calc_type == ECalcType.E_FROM_PD else _calc_rsi_from_ta


# -*- encoding:utf-8 -*-
"""
    量化波动程度模块
"""
import numpy as np
import pandas as pd
from ultron.ump.technical.line import Line
from ultron.ump.core.helper import pd_ewm_std, pd_rolling_std


def calc_wave_std(kl_pd, xd=21, ewm=True):
    """
    计算收益的移动平均std或者加权移动平均std技术线，使用
    Line封装技术线实体，不会修改kl_pd，返回UltronTLine对象
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param xd: 计算移动平均std或加权移动平均std使用的窗口参数，默认21
    :param ewm: 是否使用加权移动平均std计算
    :return: 返回Line对象
    """
    pre_close = kl_pd['close'].shift(1)
    change = np.where(pre_close == 0, 0, np.log(kl_pd['close'] / pre_close))
    if ewm:
        roll_std = pd_ewm_std(change, span=xd, min_periods=1,
                              adjust=True) * np.sqrt(xd)
    else:
        roll_std = pd_rolling_std(
            change, window=xd, min_periods=1, center=False) * np.sqrt(xd)

    # min_periods=1还是会有两个nan，填了
    roll_std = pd.Series(roll_std).fillna(method='bfill')
    line = Line(roll_std, 'wave std')
    return line


# -*- encoding:utf-8 -*-
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import ExtraTreesClassifier
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.ensemble import BaggingClassifier
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import AdaBoostRegressor
from sklearn.ensemble import AdaBoostClassifier
from sklearn.ensemble import GradientBoostingClassifier
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.tree import DecisionTreeClassifier
from sklearn.tree import DecisionTreeRegressor

from sklearn.linear_model import LinearRegression
from sklearn.linear_model import LogisticRegression
from sklearn.linear_model import Lasso
from sklearn.linear_model import SGDRegressor
from sklearn.linear_model import SGDClassifier

from sklearn.neighbors import KNeighborsClassifier
from sklearn.multiclass import OneVsOneClassifier, OneVsRestClassifier
from sklearn.svm import SVC
from xgboost import XGBClassifier
from xgboost import XGBRegressor
from lightgbm import LGBMClassifier
from lightgbm import LGBMRegressor
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import PolynomialFeatures
from ultron.ump.core.fixes import GMM

from ultron.optimize.grirdsearch import grid as MLGrid


class MLCreater(object):
    """
        默认使用线性回归初始化回归器:
            self.reg = self.linear_regression()
        默认使用带概率估计的svm初始化分类器:
            self.clf = self.svc(probability=True)

        默认无简单学习器：hmm，pca，keman全部初始None值
    """

    def __init__(self):
        self._regressor = self.linear_regression()
        self._classifier = self.svc(probability=True)

        # 无监督机器学习，pca，聚类，hmm, 默认不初始化
        self._hmm = None
        self._pca = None
        self._kmean = None

    def __str__(self):
        """打印对象显示：reg, clf, hmm, pca, kmean"""
        return "reg: {}\nclf: {}\nhmm: {}\npca: {}\nkmean: {}\n".format(
            self._regressor, self._classifier, self._hmm, self._pca,
            self._kmean)

    def linear_regression(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化LinearRegression，默认使用：
            LinearRegression()

        通过**kwargs即关键字参数透传LinearRegression，即:
            LinearRegression(**kwargs)

        :param assign: 是否保存实例后的LinearRegression对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: LinearRegression(**kwargs)
                       无参数情况下初始化: LinearRegression()

        :return: 实例化的LinearRegression对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = LinearRegression(**kwargs)
        else:
            regressor = LinearRegression()
        if assign:
            self._regressor = regressor
        return regressor

    def svc(self, assign=True, **kwargs):
        """
        有监督学习分类器，实例化SVC，默认使用：
            SVC(kernel='rbf', probability=True)

        通过**kwargs即关键字参数透传SVC，即:
            SVC(**kwargs)

        :param assign: 是否保存实例后的RandomForestRegressor对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: SVC(**kwargs)
                       无参数情况下初始化: SVC(kernel='rbf', probability=True)

        :return: 实例化的SVC对象
        """
        if kwargs is not None and len(kwargs) > 0:
            classifier = SVC(**kwargs)
        else:
            classifier = SVC(kernel='rbf', probability=True)
        if assign:
            self._classifier = classifier
        return classifier

    def pca_decomposition(self, assign=True, **kwargs):
        """
        无监督学习器，实例化PCA，默认使用pca = PCA(0.95)，通过**kwargs即
        关键字参数透传PCA，即PCA(**kwargs)

        :param assign: 是否保存实例后的PCA对象，默认True，self.pca = pca
        :param kwargs: 有参数情况下初始化: PCA(**kwargs)
                       无参数情况下初始化: pca = PCA(0.95)
        :return: 实例化的PCA对象
        """
        if kwargs is not None and len(kwargs) > 0:
            pca = PCA(**kwargs)
        else:
            # 没参数直接要保留95%
            pca = PCA(0.95)
        if assign:
            self._pca = pca

        return pca

    def kmean_cluster(self, assign=True, **kwargs):
        """
        无监督学习器，实例化KMeans，默认使用KMeans(n_clusters=2, random_state=0)，
        通过**kwargs即关键字参数透传KMeans，即KMeans(**kwargs)

        :param assign: 是否保存实例后的kmean对象，默认True，self.kmean = kmean
        :param kwargs: 有参数情况下初始化: KMeans(**kwargs)
                       无参数情况下初始化: KMeans(n_clusters=2, random_state=0)
        :return: 实例化的KMeans对象
        """
        if kwargs is not None and len(kwargs) > 0:
            kmean = KMeans(**kwargs)
        else:
            # 默认也只有两个n_clusters
            kmean = KMeans(n_clusters=2, random_state=0)
        if assign:
            self._kmean = kmean
        return kmean

    def hmm_gaussian(self, assign=True, **kwargs):
        """
        无监督学习器，实例化GMM，默认使用GMM(n_components=2)，通过**kwargs即
        关键字参数透传GMM，即GMM(**kwargs)

        导入模块使用
            try:
                from hmmlearn.hmm import GaussianHMM as GMM
            except ImportError:
                from ..CoreBu.u'l't'r'o'nFixes import GMM
        即优先选用hmmlearn中的GaussianHMM，没有安装的情况下使用sklearn中的GMM

        :param assign: 是否保存实例后的hmm对象，默认True，self.hmm = hmm
        :param kwargs: 有参数情况下初始化: GMM(**kwargs)
                       无参数情况下初始化: GMM(n_components=2)
        :return: 实例化的GMM对象
        """
        if kwargs is not None and len(kwargs) > 0:
            hmm = GMM(**kwargs)
        else:
            # 默认只有n_components=2, 两个分类
            hmm = GMM(n_components=2)
        if assign:
            self._hmm = hmm
        return hmm

    def _estimators_prarms_best(
            self,
            create_func,
            x,
            y,
            param_grid,
            assign,
            n_jobs,
            grid_callback=MLGrid.grid_search_init_n_estimators):
        """
        封装使用MLGrid寻找针对学习器的最优参数值，针对不同学习器，选择不同的
        关键字参数做最优搜索，将寻找到的最优参数做为**kwargs用来重新构造学习器

        :param create_func: callable, 学习器函数构造器，eg：self.adaboost_classifier
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，使用grid_search_mul_init_kwargs寻找参数最优值：
                        eg: _, best_params = MLGrid.grid_search_mul_init_kwargs(estimator, x, y,
                                                       param_grid=param_grid, n_jobs=n_jobs)
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True，透传create_func，用来根据最优参数重新构造学习器保存在类变量
                        eg: create_func(assign=assign, **best_params)
        :param grid_callback: 如果没有有传递最优字典关键字参数param_grid，使用学习器对应的grid_callback函数，搜索特定的最优参数
                              默认MLGrid.grid_search_init_n_estimators
        :return: 通过最优参数构造的学习器对象，eg: create_func(assign=assign, **best_params)
        """
        # 通过create_func创建一个示例学习器，assign=False
        estimator = create_func(assign=False)
        if param_grid is not None and isinstance(param_grid, dict):
            # 如果有传递最优字典关键字参数，使用grid_search_mul_init_kwargs寻找参数最优值
            grid, _, best_params = MLGrid.grid_search_mul_init_kwargs(
                estimator, x, y, param_grid=param_grid, n_jobs=n_jobs)
        else:
            # 如果没有有传递最优字典关键字参数，使用学习器对应的grid_callback函数，默认UltronMLGrid.grid_search_init_n_estimators
            grid, _, best_params = grid_callback(estimator, x, y)

        if best_params is not None:
            # 将寻找到的最优参数best_params，做为参数重新传递create_func(assign=assign, **best_params)
            self.grid = grid
            return create_func(assign=assign, **best_params)

    def bagging_classifier(self,
                           assign=True,
                           base_estimator=DecisionTreeClassifier(),
                           **kwargs):
        """
        有监督学习分类器，实例化BaggingClassifier，默认使用：
            BaggingClassifier(base_estimator=base_estimator, n_estimators=200,
                              bootstrap=True, oob_score=True, random_state=1)

        通过**kwargs即关键字参数透传BaggingClassifier，即:
            BaggingClassifier(**kwargs)

        :param base_estimator: 默认使用DecisionTreeClassifier()
        :param assign: 是否保存实例后的BaggingClassifier对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: BaggingClassifier(**kwargs)
                       无参数情况下初始化: BaggingClassifier(base_estimator=base_estimator, n_estimators=200,
                                                           bootstrap=True, oob_score=True, random_state=1)
        :return: 实例化的BaggingClassifier对象
        """
        if kwargs is not None and len(kwargs) > 0:
            if 'base_estimator' not in kwargs:
                kwargs['base_estimator'] = base_estimator
            classifier = BaggingClassifier(**kwargs)
        else:
            classifier = BaggingClassifier(base_estimator=base_estimator,
                                           n_estimators=200,
                                           bootstrap=True,
                                           oob_score=True,
                                           random_state=1)
        if assign:
            self._classifier = classifier
        return classifier

    def bagging_classifier_best(self,
                                x,
                                y,
                                param_grid=None,
                                assign=True,
                                n_jobs=-1):
        """
        寻找BaggingClassifier构造器的最优参数
        上层中bagging_classifier_best函数，直接使用ML中的x，y数据调用
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                        eg：param_grid = {'max_samples': np.arange(1, 5), 'n_estimators': np.arange(100, 300, 50)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的BaggingClassifier对象
        """
        return self._estimators_prarms_best(self.bagging_classifier, x, y,
                                            param_grid, assign, n_jobs)

    def bagging_regressor(self,
                          assign=True,
                          base_estimator=DecisionTreeRegressor(),
                          **kwargs):
        """
        有监督学习回归器，实例化BaggingRegressor，默认使用：
            BaggingRegressor(base_estimator=base_estimator, n_estimators=200,
                             bootstrap=True, oob_score=True, random_state=1)

        通过**kwargs即关键字参数透传BaggingRegressor，即:
            BaggingRegressor(**kwargs)

        :param base_estimator: 默认使用DecisionTreeRegressor()
        :param assign: 是否保存实例后的BaggingRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: BaggingRegressor(**kwargs)
                       无参数情况下初始化: BaggingRegressor(base_estimator=base_estimator, reg_core, n_estimators=200,
                                                          bootstrap=True, oob_score=True, random_state=1)
        :return: 实例化的BaggingRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            if 'base_estimator' not in kwargs:
                kwargs['base_estimator'] = base_estimator
            regressor = BaggingRegressor(**kwargs)
        else:
            regressor = BaggingRegressor(base_estimator=base_estimator,
                                         n_estimators=200,
                                         bootstrap=True,
                                         oob_score=True,
                                         random_state=1)

        if assign:
            self._regressor = regressor
        return regressor

    def bagging_regressor_best(self,
                               x,
                               y,
                               param_grid=None,
                               assign=True,
                               n_jobs=-1):
        """
        寻找BaggingRegressor构造器的最优参数
        上层UltronML中bagging_regressor_best函数，直接使用UltronML中的x，y数据调用
        
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                        eg：param_grid = {'max_samples': np.arange(1, 5), 'n_estimators': np.arange(100, 300, 50)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的BaggingRegressor对象
        """
        return self._estimators_prarms_best(self.bagging_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def adaboost_regressor(self,
                           assign=True,
                           base_estimator=DecisionTreeRegressor(),
                           **kwargs):
        """
        有监督学习回归器，实例化AdaBoostRegressor，默认使用：
            AdaBoostRegressor(base_estimator=base_estimator, n_estimators=100, random_state=1)

        通过**kwargs即关键字参数透传AdaBoostRegressor，即:
            AdaBoostRegressor(**kwargs)

        :param base_estimator: 默认使用DecisionTreeRegressor()
        :param assign: 是否保存实例后的AdaBoostRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: AdaBoostRegressor(**kwargs)
                       无参数情况下初始化: AdaBoostRegressor(n_estimators=100, random_state=1)

        :return: 实例化的AdaBoostRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            if 'base_estimator' not in kwargs:
                kwargs['base_estimator'] = base_estimator
            regressor = AdaBoostRegressor(**kwargs)
        else:
            regressor = AdaBoostRegressor(base_estimator=base_estimator,
                                          n_estimators=100,
                                          random_state=1)

        if assign:
            self._regressor = regressor

        return regressor

    def adaboost_regressor_best(self,
                                x,
                                y,
                                param_grid=None,
                                assign=True,
                                n_jobs=-1):
        """
        寻找AdaBoostRegressor构造器的最优参数

        上层UltronML中adaboost_regressor_best函数，直接使用UltronML中的x，y数据调用


        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                        eg：param_grid = {'learning_rate': np.arange(0.2, 1.2, 0.2),
                                         'n_estimators': np.arange(10, 100, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的AdaBoostRegressor对象
        """
        return self._estimators_prarms_best(self.adaboost_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def adaboost_classifier(self,
                            assign=True,
                            base_estimator=DecisionTreeClassifier(),
                            **kwargs):
        """
        有监督学习分类器，实例化AdaBoostClassifier，默认使用：
            AdaBoostClassifier(base_estimator=base_estimator, n_estimators=100, random_state=1)

        通过**kwargs即关键字参数透传AdaBoostClassifier，即:
            AdaBoostClassifier(**kwargs)

        :param base_estimator: 默认使用DecisionTreeClassifier()
        :param assign: 是否保存实例后的AdaBoostClassifier对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: AdaBoostClassifier(**kwargs)
                       无参数情况下初始化: AdaBoostClassifier(n_estimators=100, random_state=1)

        :return: 实例化的AdaBoostClassifier对象
        """
        if kwargs is not None and len(kwargs) > 0:
            if 'base_estimator' not in kwargs:
                kwargs['base_estimator'] = base_estimator
            classifier = AdaBoostClassifier(**kwargs)
        else:
            classifier = AdaBoostClassifier(base_estimator=base_estimator,
                                            n_estimators=100,
                                            random_state=1)
        if assign:
            self._classifier = classifier
        return classifier

    def adaboost_classifier_best(self,
                                 x,
                                 y,
                                 param_grid=None,
                                 assign=True,
                                 n_jobs=-1):
        """
        寻找AdaBoostClassifier构造器的最优参数

        上层UltronML中adaboost_classifier_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.2, 1.2, 0.2),
                                       'n_estimators': np.arange(10, 100, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的AdaBoostClassifier对象
        """
        return self._estimators_prarms_best(self.adaboost_classifier, x, y,
                                            param_grid, assign, n_jobs)

    def gradient_boosting_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，默认使用：
                        GradientBoostingRegressor(n_estimators=100)
        通过**kwargs即关键字参数透传GBR(**kwargs)，即:
                        GradientBoostingRegressor(**kwargs)

        :param assign: 是否保存实例后的回归器对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: GBR(n_estimators=100)
                       无参数情况下初始化: GBR(**kwargs)

        :return: 实例化的GradientBoostingRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = GradientBoostingRegressor(**kwargs)
        else:
            regressor = GradientBoostingRegressor(n_estimators=100)
        if assign:
            self._regressor = regressor
        return regressor

    def gradient_boosting_regressor_best(self,
                                         x,
                                         y,
                                         param_grid=None,
                                         assign=True,
                                         n_jobs=-1):
        """
        寻找GradientBoostingRegressor构造器的最优参数

        上层UltronML中xgb_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                      'n_estimators': np.arange(10, 100, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的GradientBoostingRegressor对象
        """
        return self._estimators_prarms_best(self.gradient_boosting_regressor,
                                            x, y, param_grid, assign, n_jobs)

    def gradient_boosting_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类器，默认使用：
                        GradientBoostingClassifier(n_estimators=100)

        通过**kwargs即关键字参数透传GradientBoostingClassifier(**kwargs)，即:
                        GradientBoostingClassifier(**kwargs)

        :param assign: 是否保存实例后的分类器对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: GradientBoostingClassifier(n_estimators=100)
                       无参数情况下初始化: GradientBoostingClassifier(**kwargs)

        :return: 实例化的GradientBoostingClassifier对象
        """
        if kwargs is not None and len(kwargs) > 0:
            classifier = GradientBoostingClassifier(**kwargs)
        else:
            classifier = GradientBoostingClassifier(n_estimators=100)
        if assign:
            self._classifier = classifier
        return classifier

    def gradient_boosting_classifier_best(self,
                                          x,
                                          y,
                                          param_grid=None,
                                          assign=True,
                                          n_jobs=-1):
        """
        寻找GradientBoostingClassifier构造器的最优参数

        上层UltronML中xgb_classifier_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                      'n_estimators': np.arange(50, 200, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的GradientBoostingClassifier对象
        """
        return self._estimators_prarms_best(self.gradient_boosting_classifier,
                                            x, y, param_grid, assign, n_jobs)

    def xgboost_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化XGBRegressor，默认使用：
            XGBRegressor(n_estimators=100)

        通过**kwargs即关键字参数透传XGBRegressor，即:
            XGBRegressor(**kwargs)

        :param assign: 是否保存实例后的XGBRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: XGBRegressor(**kwargs)
                       无参数情况下初始化: XGBRegressor(n_estimators=100)

        :return: 实例化的XGBRegressor 对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = XGBRegressor(**kwargs)
        else:
            regressor = XGBRegressor(n_estimators=100)
        if assign:
            self._regressor = regressor
        return regressor

    def xgboost_regressor_best(self,
                               x,
                               y,
                               param_grid=None,
                               assign=True,
                               n_jobs=-1):
        """
        寻找XGBRegressor构造器的最优参数

        上层UltronML中xgboost_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                      'n_estimators': np.arange(50, 200, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的XGBRegressor对象
        """
        return self._estimators_prarms_best(self.xgboost_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def xgboost_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类，实例化XGBClassifier，默认使用：
            XGBClassifier(n_estimators=100)

        通过**kwargs即关键字参数透传XGBClassifier，即:
            XGBClassifier(**kwargs)

        :param assign: 是否保存实例后的XGBClassifier对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: XGBClassifier(**kwargs)
                       无参数情况下初始化: XGBClassifier(n_estimators=100)

        :return: 实例化的XGBClassifier 对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = XGBClassifier(**kwargs)
        else:
            regressor = XGBClassifier(n_estimators=100)
        if assign:
            self._regressor = regressor
        return regressor

    def xgboost_classifier_best(self,
                                x,
                                y,
                                param_grid=None,
                                assign=True,
                                n_jobs=-1):
        """
        寻找XGBRegressor构造器的最优参数

        上层UltronML中xgboost_classifier_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                      'n_estimators': np.arange(50, 200, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的XGBClassifier对象
        """
        return self._estimators_prarms_best(self.xgboost_classifier, x, y,
                                            param_grid, assign, n_jobs)

    def lgbm_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化LGBMRegressor，默认使用：
            LGBMRegressor(n_estimators=100)

        通过**kwargs即关键字参数透传XGBRegressor，即:
            LGBMRegressor(**kwargs)

        :param assign: 是否保存实例后的LGBMRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: LGBMRegressor(**kwargs)
                       无参数情况下初始化: LGBMRegressor(n_estimators=100)

        :return: 实例化的LGBMRegressor 对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = LGBMRegressor(**kwargs)
        else:
            regressor = LGBMRegressor(n_estimators=100)
        if assign:
            self._regressor = regressor
        return regressor

    def lgbm_regressor_best(self,
                            x,
                            y,
                            param_grid=None,
                            assign=True,
                            n_jobs=-1):
        """
        寻找XGBRegressor构造器的最优参数

        上层UltronML中lgbm_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                      'n_estimators': np.arange(50, 200, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的GradientBoostingClassifier对象
        """
        return self._estimators_prarms_best(self.lgbm_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def lgbm_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类，实例化LGBMClassifier，默认使用：
            LGBMClassifier(n_estimators=100)

        通过**kwargs即关键字参数透传LGBMClassifier，即:
            LGBMClassifier(**kwargs)

        :param assign: 是否保存实例后的LGBMClassifier对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: LGBMClassifier(**kwargs)
                       无参数情况下初始化: LGBMClassifier(n_estimators=100)

        :return: 实例化的LGBMClassifier 对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = LGBMClassifier(**kwargs)
        else:
            regressor = LGBMClassifier(n_estimators=100)
        if assign:
            self._regressor = regressor
        return regressor

    def lgbm_classifier_best(self,
                             x,
                             y,
                             param_grid=None,
                             assign=True,
                             n_jobs=-1):
        """
        寻找LGBMClassifier构造器的最优参数

        上层UltronML中xgb_classifier_best函数，直接使用UltronML中的x，y数据调用
        eg：
             xgb_boost_regressor_best无param_grid参数调用：

             from Ultronpy import UltronML, ml
             ttn_Ultron = UltronML.create_test_more_fiter()
             ttn_Ultron.xgb_classifier_best()

             xgb_classifier_best有param_grid参数调用：

             param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05), 'n_estimators': np.arange(50, 200, 10)}
             ttn_Ultron.xgb_classifier_best(param_grid=param_grid, n_jobs=-1)

             out: GradientBoostingClassifier(learning_rate=0.1, n_estimators=160)

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'learning_rate': np.arange(0.1, 0.5, 0.05),
                                      'n_estimators': np.arange(50, 200, 10)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的GradientBoostingClassifier对象
        """
        return self._estimators_prarms_best(self.lgbm_classifier, x, y,
                                            param_grid, assign, n_jobs)

    def lasso_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化Lasso，默认使用：
            Lasso(n_estimators=100)

        通过**kwargs即关键字参数透传Lasso，即:
            Lasso(**kwargs)

        :param assign: 是否保存实例后的Lasso对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: Lasso(**kwargs)
                       无参数情况下初始化: Lasso(fit_intercept=True)

        :return: 实例化的Lasso对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = Lasso(**kwargs)
        else:
            regressor = Lasso(fit_intercept=True)
        if assign:
            self._regressor = regressor
        return regressor

    def lasso_regressor_best(self,
                             x,
                             y,
                             param_grid=None,
                             assign=True,
                             n_jobs=-1):
        """
        寻找RLasso构造器的最优参数

        上层UltronML中lasso_regressor_best函数，直接使用UltronML中的x，y数据调用
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(10, 150, 15)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的RandomForestRegressor对象
        """
        return self._estimators_prarms_best(self.lasso_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def sgd_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化SGDRegressor，默认使用：
            SGDRegressor(n_estimators=100)

        通过**kwargs即关键字参数透传SGDRegressor，即:
            SGDRegressor(**kwargs)

        :param assign: 是否保存实例后的SGDRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: SGDRegressor(**kwargs)
                       无参数情况下初始化: SGDRegressor(fit_intercept=True)

        :return: 实例化的SGDRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = SGDRegressor(**kwargs)
        else:
            regressor = SGDRegressor(fit_intercept=True)
        if assign:
            self._regressor = regressor
        return regressor

    def sgd_regressor_best(self,
                           x,
                           y,
                           param_grid=None,
                           assign=True,
                           n_jobs=-1):
        """
        寻找SGDRegressor构造器的最优参数

        上层UltronML中sgd_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(10, 150, 15)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的SGDRegressor对象
        """
        return self._estimators_prarms_best(self.sgd_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def sgd_classifier(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化SGDClassifier，默认使用：
            SGDClassifier(fit_intercept=True)

        通过**kwargs即关键字参数透传SGDClassifier，即:
            SGDClassifier(**kwargs)

        :param assign: 是否保存实例后的SGDClassifier对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: SGDClassifier(**kwargs)
                       无参数情况下初始化: SGDClassifier(fit_intercept=True)

        :return: 实例化的SGDClassifier对象
        """
        if kwargs is not None and len(kwargs) > 0:
            classifier = SGDClassifier(**kwargs)
        else:
            classifier = SGDClassifier(fit_intercept=True)
        if assign:
            self._classifier = classifier
        return classifier

    def sgd_classifier_best(self,
                            x,
                            y,
                            param_grid=None,
                            assign=True,
                            n_jobs=-1):
        """
        寻找SGDClassifier构造器的最优参数

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(10, 150, 15)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的SGDClassifier对象
        """
        return self._estimators_prarms_best(self.sgd_classifier, x, y,
                                            param_grid, assign, n_jobs)

    def random_forest_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化RandomForestRegressor，默认使用：
            RandomForestRegressor(n_estimators=100)

        通过**kwargs即关键字参数透传RandomForestRegressor，即:
            RandomForestRegressor(**kwargs)

        :param assign: 是否保存实例后的RandomForestRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: RandomForestRegressor(**kwargs)
                       无参数情况下初始化: RandomForestRegressor(n_estimators=100)

        :return: 实例化的RandomForestRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = RandomForestRegressor(**kwargs)
        else:
            regressor = RandomForestRegressor(n_estimators=100)
        if assign:
            self._regressor = regressor
        return regressor

    def random_forest_regressor_best(self,
                                     x,
                                     y,
                                     param_grid=None,
                                     assign=True,
                                     n_jobs=-1):
        """
        寻找RandomForestRegressor构造器的最优参数

        上层UltronML中random_forest_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(10, 150, 15)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的RandomForestRegressor对象
        """
        return self._estimators_prarms_best(self.random_forest_regressor, x, y,
                                            param_grid, assign, n_jobs)

    def random_forest_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类器，实例化RandomForestClassifier，默认使用：
            RandomForestRegressor(n_estimators=100)

        通过**kwargs即关键字参数透传RandomForestRegressor，即:
            RandomForestRegressor(**kwargs)

        :param assign: 是否保存实例后的RandomForestRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: RandomForestRegressor(**kwargs)
                       无参数情况下初始化: RandomForestRegressor(n_estimators=100)

        :return: 实例化的RandomForestRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            classifier = RandomForestClassifier(**kwargs)
        else:
            classifier = RandomForestClassifier(n_estimators=100)
        if assign:
            self._classifier = classifier
        return classifier

    def random_forest_classifier_best(self,
                                      x,
                                      y,
                                      param_grid=None,
                                      assign=True,
                                      n_jobs=-1):
        """
        寻找RandomForestClassifier构造器的最优参数

        上层UltronML中random_forest_classifier_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(50, 200, 20)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的RandomForestClassifier对象
        """
        return self._estimators_prarms_best(self.random_forest_classifier, x,
                                            y, param_grid, assign, n_jobs)

    def extra_tree_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化ExtraTreesRegressor，默认使用：
            ExtraTreesRegressor(max_depth=2, random_state=1)

        通过**kwargs即关键字参数透传ExtraTreesRegressor，即:
            ExtraTreesRegressor(**kwargs)

        :param assign: 是否保存实例后的ExtraTreesRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: ExtraTreesRegressor(**kwargs)
                       无参数情况下初始化: ExtraTreesRegressor(max_depth=2, random_state=1)

        :return: 实例化的ExtraTreesRegressor对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regressor = ExtraTreesRegressor(**kwargs)
        else:
            regressor = ExtraTreesRegressor(max_depth=2, random_state=1)
        if assign:
            self._regressor = regressor
        return regressor

    def extra_tree_regressor_best(self,
                                  x,
                                  y,
                                  param_grid=None,
                                  assign=True,
                                  n_jobs=-1):
        """
        寻找ExtraTreesRegressor构造器的最优参数

        上层UltronML中extra_tree_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(50, 200, 20)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的ExtraTreeRegressor对象
        """
        return self._estimators_prarms_best(
            self.extra_tree_regressor,
            x,
            y,
            param_grid,
            assign,
            n_jobs,
            grid_callback=MLGrid.grid_search_init_max_depth)

    def extra_tree_classifier(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化ExtraTreesClassifier，默认使用：
            ExtraTreesClassifier(max_depth=2)

        通过**kwargs即关键字参数透传ExtraTreesClassifier，即:
            ExtraTreesClassifier(**kwargs)

        :param assign: 是否保存实例后的ExtraTreesClassifier对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: ExtraTreesClassifier(**kwargs)
                       无参数情况下初始化: ExtraTreesClassifier(max_depth=2)

        :return: 实例化的ExtraTreesClassifier对象
        """
        if kwargs is not None and len(kwargs) > 0:
            classifier = ExtraTreesClassifier(**kwargs)
        else:
            classifier = ExtraTreesClassifier(max_depth=2)
        if assign:
            self._classifier = classifier
        return classifier

    def extra_tree_classifier_best(self,
                                   x,
                                   y,
                                   param_grid=None,
                                   assign=True,
                                   n_jobs=-1):
        """
        寻找ExtraTreesClassifier构造器的最优参数

        上层UltronML中decision_tree_regressor_best函数，直接使用UltronML中的x，y数据调用


        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(50, 200, 20)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的ExtraTreesClassifier对象
        """
        return self._estimators_prarms_best(
            self.extra_tree_classifier,
            x,
            y,
            param_grid,
            assign,
            n_jobs,
            grid_callback=MLGrid.grid_search_init_max_depth)

    def decision_tree_regressor(self, assign=True, **kwargs):
        """
        有监督学习回归器，实例化DecisionTreeRegressor，默认使用：
            DecisionTreeRegressor(max_depth=2, random_state=1)

        通过**kwargs即关键字参数透传DecisionTreeRegressor，即:
            DecisionTreeRegressor(**kwargs)

        :param assign: 是否保存实例后的DecisionTreeRegressor对象，默认True，self.reg = reg
        :param kwargs: 有参数情况下初始化: DecisionTreeRegressor(**kwargs)
                       无参数情况下初始化: DecisionTreeRegressor(max_depth=2, random_state=1)

        :return: 实例化的DecisionTreeRegressor对象
        """

        if kwargs is not None and len(kwargs) > 0:
            regressor = DecisionTreeRegressor(**kwargs)
        else:
            regressor = DecisionTreeRegressor(max_depth=2, random_state=1)
        if assign:
            self._regressor = regressor
        return regressor

    def decision_tree_regressor_best(self,
                                     x,
                                     y,
                                     param_grid=None,
                                     assign=True,
                                     n_jobs=-1):
        """
        寻找DecisionTreeRegressor构造器的最优参数

        上层UltronML中decision_tree_regressor_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(50, 200, 20)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的DecisionTreeRegressor对象
        """
        return self._estimators_prarms_best(
            self.decision_tree_regressor,
            x,
            y,
            param_grid,
            assign,
            n_jobs,
            grid_callback=MLGrid.grid_search_init_max_depth)

    def decision_tree_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类器，实例化DecisionTreeClassifier，默认使用：
            DecisionTreeClassifier(max_depth=2, random_state=1)

        通过**kwargs即关键字参数透传DecisionTreeClassifier，即:
            DecisionTreeClassifier(**kwargs)

        :param assign: 是否保存实例后的DecisionTreeClassifier对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: DecisionTreeClassifier(**kwargs)
                       无参数情况下初始化: DecisionTreeClassifier(max_depth=2, random_state=1)

        :return: 实例化的DecisionTreeClassifier对象
        """

        if kwargs is not None and len(kwargs) > 0:
            classifier = DecisionTreeClassifier(**kwargs)
        else:
            classifier = DecisionTreeClassifier(max_depth=2, random_state=1)
        if assign:
            self._classifier = classifier
        return classifier

    def decision_tree_classifier_best(self,
                                      x,
                                      y,
                                      param_grid=None,
                                      assign=True,
                                      n_jobs=-1):
        """
        寻找DecisionTreeClassifier构造器的最优参数

        上层UltronML中decision_tree_classifier_best函数，直接使用UltronML中的x，y数据调用

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                     eg：param_grid = {'max_features': ['sqrt', 'log2', ],
                                      'n_estimators': np.arange(50, 200, 20)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的DecisionTreeClassifier对象
        """
        return self._estimators_prarms_best(
            self.decision_tree_classifier,
            x,
            y,
            param_grid,
            assign,
            n_jobs,
            grid_callback=MLGrid.grid_search_init_max_depth)

    def knn_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类器，实例化KNeighborsClassifier，默认使用：
            KNeighborsClassifier(n_neighbors=1)

        通过**kwargs即关键字参数透传KNeighborsClassifier，即:
            KNeighborsClassifier(**kwargs)

        :param assign: 是否保存实例后的KNeighborsClassifier对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: KNeighborsClassifier(**kwargs)
                       无参数情况下初始化: KNeighborsClassifier(n_neighbors=1)

        :return: 实例化的KNeighborsClassifier对象
        """

        if kwargs is not None and len(kwargs) > 0:
            classifier = KNeighborsClassifier(**kwargs)
        else:
            classifier = KNeighborsClassifier(n_neighbors=1)
        if assign:
            self._classifier = classifier
        return classifier

    def knn_classifier_best(self,
                            x,
                            y,
                            param_grid=None,
                            assign=True,
                            n_jobs=-1):
        """
        寻找KNeighborsClassifier构造器的最优参数

        上层UltronML中knn_classifier_best函数，直接使用UltronML中的x，y数据调用
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param param_grid: 最优字典关键字参数，
                   eg：param_grid = {'algorithm': ['ball_tree', 'kd_tree', 'brute'],
                                    'n_neighbors': np.arange(1, 26, 1)}
        :param assign: 是否保存实例化后最优参数的学习器对象，默认True
        :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
        :return: 通过最优参数构造的KNeighborsClassifier对象
        """
        return self._estimators_prarms_best(
            self.knn_classifier,
            x,
            y,
            param_grid,
            assign,
            n_jobs,
            grid_callback=MLGrid.grid_search_init_n_neighbors)

    def logistic_classifier(self, assign=True, **kwargs):
        """
        有监督学习分类器，实例化LogisticRegression，默认使用：
            LogisticRegression(C=1.0, penalty='l1', tol=1e-6)

        通过**kwargs即关键字参数透传LogisticRegression，即:
            LogisticRegression(**kwargs)

        :param assign: 是否保存实例后的LogisticRegression对象，默认True，self.clf = clf
        :param kwargs: 有参数情况下初始化: LogisticRegression(**kwargs)
                       无参数情况下初始化: LogisticRegression(C=1.0, penalty='l1', tol=1e-6)

        :return: 实例化的LogisticRegression对象
        """
        if kwargs is not None and len(kwargs) > 0:
            regression = LogisticRegression(**kwargs)
        else:
            regression = LogisticRegression(C=1.0, penalty='l1', tol=1e-6)
        if assign:
            self._regressor = regression
        return regression

    def polynomial_regression(self, assign=True, degree=2, **kwargs):
        """
        有监督学习回归器，使用：
            make_pipeline(PolynomialFeatures(degree), LinearRegression(**kwargs))

        :param assign: 是否保存实例后的LinearRegression对象，默认True，self.reg = reg
        :param degree: 多项式拟合参数，默认2
        :param kwargs: 由make_pipeline(PolynomialFeatures(degree), LinearRegression(**kwargs))
                       即关键字参数**kwargs全部传递给LinearRegression做为构造参数

        :return: 实例化的回归对象
        """
        regression = make_pipeline(PolynomialFeatures(degree),
                                   LinearRegression(**kwargs))
        if assign:
            self._regressor = regression
        return regression

    def onevsone_classifier(self, assign=False, **kwargs):
        """
        封装有监督学习分类器，使用OneVsOneClassifier进行多label的
        分类器二次封装，即：
             OneVsOneClassifier(self.clf, **kwargs)

        :param assign: 是否保存实例后的二次封装分类器对象，与其它构造器不同，
                       默认False，即默认不保存在类中替换原始分类器
        :param kwargs: 透传OneVsOneClassifier做为构造关键字参数
        :return: OneVsOneClassifier对象
        """
        onevsone = OneVsOneClassifier(self._classifier, **kwargs)
        if assign:
            self._classifier = onevsone
        return onevsone

    def onevsreset_classifier(self, assign=False, **kwargs):
        """
        封装有监督学习分类器，使用OneVsRestClassifier进行多label的
        分类器二次封装，即：
             OneVsRestClassifier(self.clf, **kwargs)

        :param assign: 是否保存实例后的二次封装分类器对象，与其它构造器不同，
                       默认False，即默认不保存在类中替换原始分类器
        :param kwargs: 透传OneVsRestClassifier做为构造关键字参数
        :return: OneVsRestClassifier对象
        """
        onevsreset = OneVsRestClassifier(self._classifier, **kwargs)
        if assign:
            self._classifier = onevsreset
        return onevsreset


# -*- encoding:utf-8 -*-
"""
    计算线atr模块
"""

import numpy as np
import pandas as pd
from ultron.ump.core.helper import pd_ewm_std, pd_rolling_std
from ultron.ump.technical.line import Line


def calc_atr_std(kl_pd, xd=21, ewm=True):
    """
    计算atr移动平均std或者加权移动平均std技术线，使用
    UltronTLine封装技术线实体，不会修改kl_pd，返回UltronTLine对象
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param xd: 计算移动平均std或加权移动平均std使用的窗口参数，默认21
    :param ewm: 是否使用加权移动平均std计算
    :return: 返回UltronTLine对象
    """
    pre_atr21 = kl_pd['atr21'].shift(1)
    # noinspection PyTypeChecker
    atr_change = np.where(pre_atr21 == 0, 0,
                          np.log(kl_pd['atr21'] / pre_atr21))

    if ewm:
        atr_roll_std = pd_ewm_std(
            atr_change, span=xd, min_periods=1, adjust=True) * np.sqrt(xd)
    else:
        atr_roll_std = pd_rolling_std(
            atr_change, window=xd, min_periods=1, center=False) * np.sqrt(xd)

    # min_periods=1还是会有两个nan，填了
    atr_roll_std = pd.Series(atr_roll_std).fillna(method='bfill')
    # 主要目的就是通过atr_roll_std构造line
    line = Line(atr_roll_std, 'atr std')
    return line


# -*- coding: utf-8 -*-
import json
import base64
import os
import shutil
from ultron.utilities.short_uuid import uuid, decode
from ultron.cluster.work.extern_modules.base_module import BaseModule
from ultron.utilities.zlib_engine import unzip_compress


class Module(BaseModule):

    def __init__(self, name, wid, token, redis_client):
        super(Module, self).__init__(name, wid, token, redis_client)
        self._func = {'upload_packet': self.upload_packet}
        self._namespace = 'packet'

    def process_respone(self, result):
        name = result['name']
        opcode = result['opcode']
        self._func[opcode](result)

    def upload_packet(self, result):
        uid = result['uid']
        dir_name = './'
        file_info = bytes(result['file_info'], encoding='utf-8')
        packet_name = result['packet_name']
        file_name = result['file_name']
        all_dir_name = os.path.join(dir_name, str(uid))
        file_list = []
        if not os.path.exists(all_dir_name):
            os.makedirs(all_dir_name)
        file_content = base64.b64decode(file_info)

        #检查目录是否存在
        if os.path.exists(os.path.join(all_dir_name, str(packet_name))):
            shutil.rmtree(os.path.join(all_dir_name, str(packet_name)))

        #重新创建
        os.mkdir(os.path.join(all_dir_name, str(packet_name)))

        #保存成文件
        new_file = os.path.join(all_dir_name, str(packet_name), file_name)
        with open(new_file, 'wb') as f:
            f.write(file_content)
        unzip_compress(new_file, os.path.join(all_dir_name, str(packet_name)))
        if os.path.isfile(new_file):
            os.remove(new_file)

        ## 通知主节点发布成功，进行启动
        packet_info = {
            'name': 'tasks',
            'opcode': 'startup_task',
            'dir_name': all_dir_name,
            'wid': self._wid,
            'work_name': packet_name,
            'token': self._token
        }
        self._redis_client.hset('ultron:work:ctask', uuid(),
                                json.dumps(packet_info))



# -*- encoding:utf-8 -*-
"""
    类基础通用模块
"""
import sys
import pandas as pd
from ultron import __version__
from ultron.ump.core.fixes import signature, Parameter, pickle


class FreezeAttrMixin(object):
    """冻结对外设置属性混入类，设置抛异常"""

    def _freeze(self):
        """冻结属性设置接口"""
        object.__setattr__(self, "__frozen", True)

    def __setattr__(self, key, value):
        if getattr(
                self, "__frozen",
                False) and not (key in type(self).__dict__ or key == "_cache"):
            raise AttributeError(
                "You cannot add any new attribute '{key}'".format(key=key))
        object.__setattr__(self, key, value)


class PickleStateMixin(object):
    """混入有本地序列化需求的类"""

    # pickle的最高支持版本
    _pickle_highest_protocol = pickle.HIGHEST_PROTOCOL
    # python版本，也可简单使用是否py3
    _python_version = str(sys.version_info)
    # windows or mac os
    _is_mac_os = False  #ABuEnv.g_is_mac_os
    # 是否考虑本身的版本version, 默认不考虑，忽略abupy的版本号
    skip_ultron_version = True

    def __getstate__(self):
        _ultron_version = __version__
        self.pick_extend_work()
        return dict(self.__dict__.items(),
                    _ultron_version=_ultron_version,
                    _pickle_highest_protocol=self._pickle_highest_protocol,
                    _python_version=self._python_version,
                    _is_mac_os=self._is_mac_os)

    def __setstate__(self, state):
        """开始从本地序列化文件转换为python对象，即unpick"""

        # 从本地序列化文件中读取的pickle的最高支持版本, 默认0
        pickle_highest_protocol = state.pop("_pickle_highest_protocol", 0)
        # 从本地序列化文件中读取的abupy的版本号, 默认0.0.1
        old_ultron_version = state.pop("_ultron_version", '0.0.1')
        # 从本地序列化文件中读取的python版本号, 默认2.7.0
        python_version = state.pop("_python_version", '2.7.0')
        # 从本地序列化文件中读取的平台信息, 默认False，即windows
        platform_version = state.pop("_is_mac_os", False)

        if self.skip_ultron_version:
            # 忽略abupy的版本号
            _ultron_version = old_ultron_version
        else:
            _ultron_version = __version__

        if self._pickle_highest_protocol != pickle_highest_protocol \
                or _ultron_version != old_ultron_version or self._python_version != python_version \
                or self._is_mac_os != platform_version:
            """只要有一个信息不一致，打印info，即有序列化读取失败的可能"""
            print("unpickle {} : "
                  "old pickle_highest_protocol={},"
                  "now pickle_highest_protocol={}, "
                  "old abupy_version={}, "
                  "now abupy_version={}, "
                  "old python_version={}, "
                  "now python_version={}, "
                  "old platform_version={}, "
                  "now platform_version={}, ".format(
                      self.__class__.__name__, pickle_highest_protocol,
                      self._pickle_highest_protocol, old_ultron_version,
                      _ultron_version, python_version, self._python_version,
                      platform_version, self._is_mac_os))

        self.__dict__.update(state)
        # 混入对象可覆盖unpick_extend_work方法，完成对象特有的unpick工作
        self.unpick_extend_work(state)

    # noinspection PyMethodMayBeStatic
    def pick_extend_work(self):
        """混入对象可覆盖pick_extend_work方法，完成对象特有的__getstate__工作"""
        pass

    def unpick_extend_work(self, state):
        """混入对象可覆盖unpick_extend_work方法，完成对象特有的__setstate__工作"""
        pass


class ParamBase(object):
    """对象基础类，实现对象基本信息打印，调试查看接口"""

    @classmethod
    def get_params(cls):
        # init中特意找了被类装饰器替换了的deprecated_original方法，即原始init方法
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # 非自定义init返回空
            return list()
        # 获取init的参数签名
        init_signature = signature(init)
        # 过滤self和func(*args), 和func(**kwargs)
        parameters = [
            p for p in init_signature.parameters.values()
            if p.name != 'self' and p.kind != Parameter.VAR_KEYWORD
            and p.kind != Parameter.VAR_POSITIONAL
        ]
        return sorted([p.name for p in parameters])

    def _filter_attr(self, user):
        """根据user设置，返回所有类属性key或者用户定义类属性key"""
        if not user:
            return self.__dict__.keys()

        # 只筛选用户定义类属性key
        user_attr = list(
            filter(lambda attr: not attr.startswith('_'),
                   self.__dict__.keys()))
        return user_attr

    def to_dict(self, user=True):
        """for debug show dict"""
        return {attr: self.__dict__[attr] for attr in self._filter_attr(user)}

    def to_series(self, user=True):
        """for notebook debug show series"""
        return pd.Series(self.to_dict(user))

    def __str__(self):
        """打印对象显示：class name, params"""
        class_name = self.__class__.__name__
        return '%s(%s)' % (class_name, self.get_params())

    __repr__ = __str__



# -*- encoding:utf-8 -*-
"""
    期货symbol数据模块
"""

import os

import pandas as pd
from ultron.env import g_base_path
from ultron.ump.core.base import FreezeAttrMixin
from ultron.ump.core import env
from ultron.kdutils.lazy import LazyFunc
from ultron.kdutils.decorator import singleton
from ultron.ump.market.symbol import Symbol
"""国内期货symbol文件，文件定期重新爬取，更新"""

_stock_code_futures_cn = os.path.join(g_base_path, 'futures_cn.csv')


@singleton
class FuturesCn(FreezeAttrMixin):
    """国内期货symbol数据，AbuFuturesCn单例，混入FreezeAttrMixin在__init__中冻结了接口，外部只可以读取"""

    def __init__(self):
        """
            self.futures_cn_df表结构如下所示：列分别为
            symbol product	min_deposit	min_unit	commission	exchange

                0	V0	PVC	0.07	5	2.0	DCE
                1	P0	棕榈	0.07	10	2.5	DCE
                2	B0	豆二	0.05	10	2.0	DCE
                3	M0	豆粕	0.07	10	1.5	DCE
                4	I0	铁矿石	0.10	100	8.0	DCE
                5	JD0 	鸡蛋	0.08	5	6.0	DCE
                6	L0	塑料	0.07	5	2.0	DCE
                7	PP0	PP	0.07	5	4.0	DCE
                8	FB0	纤维板	0.20	500	10.0	DCE
                9	BB0	胶合板	0.20	500	10.0	DCE
                10	Y0	豆油	0.07	10	2.5	DCE
            如第5行：symbol＝JD0，product＝鸡蛋，min_deposit(最少保证金比例)＝0.08
                    min_unit(每手最少交易单位)=5, commission(每一手手续费)=6.0,
                    exchange(交易所)=DCE
        """
        # 读取本地csv到内存
        self.futures_cn_df = pd.read_csv(_stock_code_futures_cn, index_col=0)
        # __init__中使用FreezeAttrMixin._freeze冻结了接口
        self._freeze()

    def __str__(self):
        """打印对象显示：futures_cn_df.info， futures_cn_df.describe"""
        return 'info:\n{}\ndescribe:\n{}'.format(self.futures_cn_df.info(),
                                                 self.futures_cn_df.describe())

    __repr__ = __str__

    def __len__(self):
        """对象长度：futures_cn_df.shape[0]，即行数"""
        return self.futures_cn_df.shape[0]

    def __contains__(self, item):
        """成员测试：item in self.futures_cn_df.columns，即item在不在futures_cn_df的列中"""
        return item in self.futures_cn_df.columns

    def __getitem__(self, key):
        """索引获取：套接self.futures_cn_df[key]"""
        if key in self:
            return self.futures_cn_df[key]
        # 不在的话，返回整个表格futures_cn_df

        symbol_df = self.futures_cn_df[self.futures_cn_df.symbol == key]
        if not symbol_df.empty:
            return symbol_df
        return self.futures_cn_df

    def __setitem__(self, key, value):
        """索引设置：对外抛出错误， 即不准许外部设置"""
        raise AttributeError("AbuFuturesCn set value!!!")

    def query_symbol(self, symbol):
        """
        对外查询接口
        :param symbol: 可以是Symbol对象，也可以是symbol字符串对象
        """
        if isinstance(symbol, Symbol):
            symbol = symbol.value

        if symbol[0].isalpha() and symbol[1:].isdigit():
            head = symbol[0]
        elif symbol[:2].isalpha() and symbol[2:].isdigit():
            head = symbol[:2]
        else:
            return None

        for fs in self.symbol:
            if head == fs[:len(head)]:
                return self.futures_cn_df[self.futures_cn_df.symbol == fs]
        return None

    def query_min_unit(self, symbol):
        """
        对query_symbol的查询一手单位的近一部函数封装
        :param symbol: 可以是Symbol对象，也可以是symbol字符串对象
        """
        min_cnt = 10
        # 查询最少一手单位
        q_df = self.query_symbol(symbol)
        if q_df is not None:
            min_cnt = q_df.min_unit.values[0]
        return min_cnt

    @LazyFunc
    def symbol(self):
        """代理获取futures_cn_df.symbol，LazyFunc"""
        return self.futures_cn_df.symbol

    @LazyFunc
    def product(self):
        """代理获取futures_cn_df.product，LazyFunc"""
        return self.futures_cn_df.product

    @LazyFunc
    def min_deposit(self):
        """代理获取futures_cn_df.min_deposit，LazyFunc"""
        return self.futures_cn_df.min_deposit

    @LazyFunc
    def min_unit(self):
        """代理获取futures_cn_df.min_unit，LazyFunc"""
        return self.futures_cn_df.min_unit

    @LazyFunc
    def commission(self):
        """代理获取futures_cn_df.commission，LazyFunc"""
        return self.futures_cn_df.commission

    @LazyFunc
    def exchange(self):
        """代理获取futures_cn_df.exchange，LazyFunc"""
        return self.futures_cn_df.exchange



# -*- encoding:utf-8 -*-
"""
    交易执行模块
"""
import numpy as np
import pandas as pd
from ultron.utilities.logger import kd_logger
from ultron.kdutils.progress import Progress
from ultron.kdutils import date as date_utils
from ultron.ump.trade.ml_feature import MlFeature


def calc_simple_profit(orders, kl_pd):
    """
    计算交易收益，simple的意思是不考虑手续费
    :param orders: Order对象序列
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :return:
    """
    all_profit = 0
    now_price = kl_pd[-1:].close
    for order in orders:
        if order.sell_type == 'keep':
            # 单子如果还没有卖出，使用now_price计算收益
            all_profit += (now_price - order.buy_price
                           ) * order.buy_cnt * order.expect_direction
        else:
            # 单子如卖出，使用sell_price计算收益
            all_profit += (order.sell_price - order.buy_price
                           ) * order.buy_cnt * order.expect_direction
    return all_profit


def make_orders_pd(orders, kl_pd):
    """
    AbuOrder对象序列转换为pd.DataFrame对象，order_pd中每一行代表一个AbuOrder信息
    :param orders: AbuOrder对象序列
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    """
    ret_orders_pd = None
    for index, order in enumerate(orders):
        # 迭代order，将每一个AbuOrder对象转换为一个pd.DataFrame对象
        order_pd = pd.DataFrame(np.array([
            order.buy_date, order.buy_price, order.buy_cnt, order.buy_factor,
            order.buy_symbol, order.buy_pos, order.buy_type_str,
            order.expect_direction, order.sell_type_extra, order.sell_date,
            order.sell_price, order.sell_type, order.ml_features
        ]).reshape(1, -1),
                                index=[index],
                                columns=[
                                    'buy_date', 'buy_price', 'buy_cnt',
                                    'buy_factor', 'symbol', 'buy_pos',
                                    'buy_type_str', 'expect_direction',
                                    'sell_type_extra', 'sell_date',
                                    'sell_price', 'sell_type', 'ml_features'
                                ])

        # 从原始金融时间序列中找到key，赋予order_pd['key']
        mask = kl_pd[kl_pd['date'] == order.buy_date]
        order_pd['key'] = mask['key'].values[0]
        # 将所有order_pd concat生成一个pd.DataFrame对象
        ret_orders_pd = order_pd if ret_orders_pd is None else pd.concat(
            [ret_orders_pd, order_pd])

    # 转换连接好的pd.DataFrame对象的index赋予对应的时间，形成交易时间序列
    dates_fmt = list(
        map(lambda date: date_utils.fmt_date(date),
            ret_orders_pd['buy_date'].tolist()))
    dates_pd = pd.to_datetime(dates_fmt)
    ret_orders_pd.index = dates_pd

    # 把除字符串类型外的所有进行列类型进行显示转换，因为支持py3
    ret_orders_pd['sell_price'] = ret_orders_pd['sell_price'].astype(float)
    ret_orders_pd['sell_date'] = ret_orders_pd['sell_date'].fillna(0).astype(
        int)

    ret_orders_pd['buy_price'] = ret_orders_pd['buy_price'].astype(float)
    ret_orders_pd['buy_date'] = ret_orders_pd['buy_date'].astype(int)
    ret_orders_pd['buy_cnt'] = ret_orders_pd['buy_cnt'].astype(float)
    ret_orders_pd['expect_direction'] = ret_orders_pd[
        'expect_direction'].astype(float)

    # 计算收益
    c_ss = (ret_orders_pd['sell_price'] - ret_orders_pd['buy_price']
            ) * ret_orders_pd['buy_cnt'] * ret_orders_pd['expect_direction']
    ret_orders_pd['profit'] = np.round(c_ss.values, decimals=2)

    # 判定单子最终是否盈利 win：1，loss：－1. keep：0
    # noinspection PyTypeChecker
    ret_orders_pd['result'] = np.where(ret_orders_pd['sell_type'] == 'win', 1,
                                       -1)
    # 针对还是keep的单子置0
    # noinspection PyTypeChecker
    ret_orders_pd['result'] = np.where(ret_orders_pd['sell_type'] == 'keep', 0,
                                       ret_orders_pd['result'])
    # 如果单子开启了特征收集，将收集的特征添加到对应的交易中，详阅读MlFeature
    MlFeature().unzip_ml_feature(ret_orders_pd)
    return ret_orders_pd


def transform_action(orders_pd):
    """
    将在make_orders_pd中交易订单构成的pd.DataFrame对象进行拆解，分成买入交易行为及数据，卖出交易行为和数据，
    按照买卖时间顺序，转换构造交易行为顺序序列
    :param orders_pd: 交易订单构成的pd.DataFrame对象
    :return: 交易行为顺序序列 pd.DataFrame对象
    """

    # 从order中摘出买入交易行为
    buy_actions = orders_pd.loc[:, [
        'buy_date', 'buy_price', 'buy_cnt', 'symbol', 'expect_direction',
        'sell_price'
    ]]
    # action = buy
    buy_actions['action'] = 'buy'
    # ACTION和order都有的action使用首字母大写，内容小写区分开
    buy_actions = buy_actions.rename(
        columns={
            'buy_date': 'Date',
            'buy_price': 'Price',
            'buy_cnt': 'Cnt',
            'sell_price': 'Price2',
            'expect_direction': 'Direction'
        })
    buy_actions.index = np.arange(0, buy_actions.shape[0])

    # 从order中摘出卖出交易行为
    sell_actions = orders_pd.loc[:, [
        'sell_date', 'sell_price', 'buy_cnt', 'symbol', 'expect_direction',
        'buy_price'
    ]]
    # action = sell
    sell_actions['action'] = 'sell'
    # action和order都有的action使用首字母大写，内容小写区分开
    sell_actions = sell_actions.rename(
        columns={
            'sell_date': 'Date',
            'sell_price': 'Price',
            'buy_cnt': 'Cnt',
            'buy_price': 'Price2',
            'expect_direction': 'Direction'
        })
    sell_actions.index = np.arange(
        buy_actions.shape[0], buy_actions.shape[0] + sell_actions.shape[0])

    # 把买入交易行为和卖出交易行为连起来
    action_pd = pd.concat([buy_actions, sell_actions])

    # 根据时间和买卖行为排序，即构成时间行为顺序
    # noinspection PyUnresolvedReferences
    action_pd = action_pd.sort_values(['Date', 'action'])
    action_pd.index = np.arange(0, action_pd.shape[0])
    # action中干掉所有keep的单子, 只考虑Price列，即drop卖出行为Price是nan的
    action_pd = action_pd.dropna(subset=['Price'])
    # 一定要先把date转换成int sort_values
    action_pd['Date'] = action_pd['Date'].astype(int)
    action_pd = action_pd.sort_values(['Date', 'action'])
    return action_pd


def apply_action_to_capital(capital,
                            action_pd,
                            kl_pd_manager,
                            show_progress=True):
    """
    多个金融时间序列对应的多个交易行为action_pd，在考虑资金类AbuCapital对象的情况下，对AbuCapital对象进行
    资金时间序列更新，以及判定在有限资金的情况下，交易行为是否可以执行
    :param capital: 资金类AbuCapital实例化对象
    :param action_pd: 交易行为构成的pd.DataFrame对象
    :param kl_pd_manager: 金融时间序列管理对象，AbuKLManager实例
    :param show_progress: 是否显示进度条，默认True
    :return:
    """
    if action_pd.empty:
        kd_logger.info('apply_action_to_capital action_pd.empty!!!')
        return

    # 如果交易symbol数量 > 10000个显示初始化进度条
    init_show_progress = (show_progress and len(set(action_pd.symbol)) > 10000)
    # 资金时间序列初始化各个symbol对应的持仓列，持仓价值列
    capital.apply_init_kl(action_pd, show_progress=init_show_progress)

    # 如果交易symbol数量 > 1个显示apply进度条
    show_apply_act_progress = (show_progress
                               and len(set(action_pd.symbol)) > 1)
    # 外部new一个进度条，为使用apply的操作使用
    with Progress(len(action_pd), 0, label='capital.apply_action') as progress:
        # 针对每一笔交易进行buy，sell细节处理，涉及有限资金是否成交判定
        action_pd['deal'] = action_pd.apply(
            capital.apply_action,
            axis=1,
            args=(progress if show_apply_act_progress else None, ))

    # 如果交易symbol数量 > 1000个显示apply进度条
    show_apply_kl = (show_progress and len(set(action_pd.symbol)) > 1000)
    # 根据交易行为产生的持仓列，持仓价值列更新资金时间序列
    capital.apply_kl(action_pd, kl_pd_manager, show_progress=show_apply_kl)

    # filter出所有持仓价值列
    stock_worths = capital.capital_pd.filter(regex='.*_worth')
    # 所有持仓价值列的sum形成stocks_blance列
    capital.capital_pd['stocks_blance'] = stock_worths.sum(axis=1)
    # stocks_blance ＋ cash_blance（现金余额）＝ capital_blance（总资产价值）列
    capital.capital_pd['capital_blance'] = capital.capital_pd[
        'stocks_blance'] + capital.capital_pd['cash_blance']



# -*- encoding:utf-8 -*-
"""
    跳空缺口模块
"""
import numpy as np
import pandas as pd
from collections import namedtuple
from ultron.ump.core.helper import pd_resample
from ultron.kdutils.date import fmt_date


def calc_jump(kl_pd, jump_diff_factor=1):
    """
    通过对比交易日当月的成交量，和当月的振幅来确定交易日当日的跳空阀值，
    分别组装跳空方向，跳空能量，跳空距离等数据进入pd.DataFrame对象返回
    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param jump_diff_factor: 参数通过设置jump_diff_factor来调节跳空阀值的大小，默认jump_diff_factor＝1
    :param show: 是否对结果跳空点进行可视化
    :return: pd.DataFrame对象
    """
    # 由于过程会修改金融时间序列，所以先copy一个再处理
    kl_pd = kl_pd.copy()
    # 将日change取abs变成日振幅保存在kl_pd新列abs_pct_change
    kl_pd['abs_pct_change'] = np.abs(kl_pd['p_change'])
    # 日振幅取平均做为第一层判断是否达成跳空的条件，即跳空最起码要振幅超过日振幅平均值
    change_ratio_min = kl_pd['abs_pct_change'].mean()

    # 提取月振幅volume_mean
    # TODO 做为参数可修改21d
    change_mean = pd_resample(kl_pd.abs_pct_change, '21D', how='mean')
    """
        eg: change_mean形如
        2014-07-23    0.7940
        2014-08-13    0.6536
        2014-09-03    0.8120
        2014-09-24    1.2673
        2014-10-15    1.1007
                       ...
        2016-04-13    1.2080
        2016-05-04    0.9093
        2016-05-25    0.6208
        2016-06-15    1.1831
        2016-07-06    0.6693
    """
    # 提取月成交量均值volume_mean
    volume_mean = pd_resample(kl_pd.volume, '21D', how='mean')
    """
        eg：volume_mean形如
        2014-07-23    1350679
        2014-08-13    1256093
        2014-09-03    1593358
        2014-09-24    1816544
        2014-10-15    2362897
                       ...
        2016-04-13    2341972
        2016-05-04    1633200
        2016-05-25    1372525
        2016-06-15    2071612
        2016-07-06    1136278
    """
    # 使用使用kl_pd没有resample之前的index和change_mean进行loc操作，为了把没有的index都变成nan
    change_mean = change_mean.reindex(kl_pd.index)
    change_mean = change_mean.loc[kl_pd.index.intersection(change_mean.index)]
    # 有nan之后开始填充nan
    change_mean.fillna(method='pad', inplace=True)
    # bfill再来一遍只是为了填充最前面的nan
    change_mean.fillna(method='bfill', inplace=True)
    """
        loc以及填充nan后change_mean形如：change_mean
        2014-07-23    0.7940
        2014-07-24    0.7940
        2014-07-25    0.7940
        2014-07-28    0.7940
        2014-07-29    0.7940
        2014-07-30    0.7940
        2014-07-31    0.7940
        2014-08-01    0.7940
        2014-08-04    0.7940
        2014-08-05    0.7940
                       ...
        2016-07-13    0.6693
        2016-07-14    0.6693
        2016-07-15    0.6693
        2016-07-18    0.6693
        2016-07-19    0.6693
        2016-07-20    0.6693
        2016-07-21    0.6693
        2016-07-22    0.6693
        2016-07-25    0.6693
        2016-07-26    0.6693
    """
    # 使用使用kl_pd没有resample之前的index和change_mean进行loc操作，为了把没有的index都变成nan
    volume_mean = volume_mean.reindex(kl_pd.index)
    volume_mean = volume_mean.loc[kl_pd.index.intersection(volume_mean.index)]
    # 有nan之后开始填充nan
    volume_mean.fillna(method='pad', inplace=True)
    # bfill再来一遍只是为了填充最前面的nan
    volume_mean.fillna(method='bfill', inplace=True)
    """
        loc以及填充nan后volume_mean形如：change_mean
        2014-07-23    1350679.0
        2014-07-24    1350679.0
        2014-07-25    1350679.0
        2014-07-28    1350679.0
        2014-07-29    1350679.0
        2014-07-30    1350679.0
        2014-07-31    1350679.0
        2014-08-01    1350679.0
        2014-08-04    1350679.0
        2014-08-05    1350679.0
                        ...
        2016-07-13    1136278.0
        2016-07-14    1136278.0
        2016-07-15    1136278.0
        2016-07-18    1136278.0
        2016-07-19    1136278.0
        2016-07-20    1136278.0
        2016-07-21    1136278.0
        2016-07-22    1136278.0
        2016-07-25    1136278.0
        2016-07-26    1136278.0
    """
    jump_pd = pd.DataFrame()

    # 迭代金融时间序列，即针对每一个交易日分析跳空
    for kl_index in np.arange(0, kl_pd.shape[0]):
        today = kl_pd.iloc[kl_index]
        if today.abs_pct_change <= change_ratio_min:
            # 第一层判断：跳空最起码要振幅超过日振幅平均值
            continue
        date = fmt_date(today.date)
        if date not in volume_mean.index:
            continue

        if today.volume <= volume_mean.loc[date]:
            # 第二层判断：跳空当日的成交量起码要超过当月平均值
            continue

        # 获取今天对应的月振幅, 做为今天判断是否跳空的价格阀值百分比
        jump_threshold = np.abs(change_mean.loc[date])
        if today.pre_close == 0 or jump_threshold == 0:
            # 只是为避免异常数据
            continue

        # 计算跳空距离阀值，即以昨天收盘为基数乘以跳空阀值比例除100得到和高开低收相同单位的价格阀值jump_diff
        # 参数通过设置jump_diff_factor来调节跳空阀值的大小，默认jump_diff_factor＝1
        jump_diff = today.pre_close * jump_threshold / 100 * jump_diff_factor
        # 第三层判断：根据向上向下跳空选择跳空计算
        if today.p_change > 0 and (today.low - today.pre_close) > jump_diff:
            # 注意向上跳空判断使用today.low，向上跳空 1
            today['jump'] = 1
            # 月振幅跳空阀值
            today['jump_threshold'] = jump_threshold
            # 跳空距离阀值
            today['jump_diff'] = jump_diff
            # 计算出跳空缺口强度
            today['jump_power'] = (today.low - today.pre_close) / jump_diff

            jump_pd = jump_pd.append(today)
        elif today.p_change < 0 and (today.pre_close - today.high) > jump_diff:
            # 注意向下跳空判断使用today.high，向下跳空 －1
            today['jump'] = -1
            # 月振幅跳空阀值
            today['jump_threshold'] = jump_threshold
            # 跳空距离阀值
            today['jump_diff'] = jump_diff
            # 计算出跳空缺口强度
            today['jump_power'] = (today.pre_close - today.high) / jump_diff
            jump_pd = jump_pd.append(today)
    return jump_pd


# noinspection PyClassHasNoInit
class JumpTuple(
        namedtuple('JumpTuple', ('date', 'direction', 'power', 'price'))):
    __slots__ = ()

    def __repr__(self):
        return "date:{}, direction:{}, power:{}, price:{}".format(
            self.date, self.direction, self.power, self.price)


def calc_jump_line(kl_pd, power_threshold=2.0, jump_diff_factor=1):
    """
    通过calc_jump计算kl_pd金融时间序列周期内跳空方向，跳空能量，跳空距离，
    通过跳空能量jump.jump_power大于阀值的跳空点组成AbuJumpTuple对象，
    AbuJumpTuple对象有跳空日期date，跳空方向direction，跳空能量power，跳空基数价格price组成，
    可视化AbuJumpTuple对象序列jump_lines中所有的跳空点

    :param kl_pd: 金融时间序列，pd.DataFrame对象
    :param power_threshold: 筛选跳空的点阀值，float，默认2.0
    :param jump_diff_factor: 透传给calc_jump，参数通过设置jump_diff_factor来调节跳空阀值的大小
    :return: AbuJumpTuple对象序列
    """
    # 通过calc_jump计算kl_pd金融时间序列周期内跳空方向，跳空能量，跳空距离pd.DataFrame对象jump_pd
    jump_pd = calc_jump(kl_pd, jump_diff_factor=jump_diff_factor)
    jump_lines = []
    for jump_index in np.arange(0, jump_pd.shape[0]):
        jump = jump_pd.iloc[jump_index]
        # 通过跳空能量jump.jump_power大于阀值的跳空点组成AbuJumpTuple对象
        if jump.jump_power > power_threshold:
            # AbuJumpTuple对象有跳空日期date，跳空方向direction，跳空能量power，跳空基数价格price组成
            jump_tuple = JumpTuple(jump_pd.index[jump_index], jump.jump,
                                   jump.jump_power, jump.pre_close)
            jump_lines.append(jump_tuple)
    return jump_lines


def calc_jump_line_weight(kl_pd,
                          sw=(0.5, 0.5),
                          power_threshold=2.0,
                          jump_diff_factor=1):
    """
    通过calc_jump计算kl_pd金融时间序列周期内跳空方向，跳空能量，跳空距离，
    把每一个跳空点都转换为AbuJumpTuple对象，与calc_jump_line不同点是
    计算时间跳空能量分两部分组成 非时间加权的跳空能量 ＋ 时间加权的跳空能量，
    参数sw控制占比，sw[0]：控制非时间加权的跳空能量所最终占比，sw[1]：控制时间加权的跳空能量所最终占比
    最终加权能量大于阀值能量的进行能量替换，加入到结果序列中返回
    :param kl_pd:  金融时间序列，pd.DataFrame对象
    :param sw: tuple对象，sw[0]：控制非时间加权的跳空能量所最终占比，sw[1]：控制时间加权的跳空能量所最终占比
    :param power_threshold: 筛选跳空的点阀值，float，默认2.0
    :param jump_diff_factor: 透传给calc_jump，参数通过设置jump_diff_factor来调节跳空阀值的大小
    :return: AbuJumpTuple对象序列
    """
    # 通过calc_jump计算kl_pd金融时间序列周期内跳空方向，跳空能量，跳空距离pd.DataFrame对象jump_pd
    jump_pd = calc_jump(kl_pd, jump_diff_factor=jump_diff_factor)
    ws_jump_lines = []
    for jump_index in np.arange(0, jump_pd.shape[0]):
        jump = jump_pd.iloc[jump_index]
        # 把每一个跳空点都转换为AbuJumpTuple
        jump_tuple = JumpTuple(jump_pd.index[jump_index], jump.jump,
                               jump.jump_power, jump.pre_close)
        # 拿出跳空日的index，之后拿对应的跳空权重值使用
        weight_index = kl_pd.index.tolist().index(jump_tuple.date)
        # 线性加权0-1kl_pd.shape[0]个, 之后对应时间序列
        weights = np.linspace(0, 1, kl_pd.shape[0])
        """
        eg: weights 形如
        array([ 0.    ,  0.002 ,  0.004 ,  0.006 ,  0.008 ,  0.0099,  0.0119,
                0.0139,  0.0159,  0.0179,  0.0199,  0.0219,  0.0239,  0.0258,
                0.0278,  0.0298,  0.0318,  0.0338,  0.0358,  0.0378,  0.0398,
                0.0417,  0.0437,  0.0457,  0.0477,  0.0497,  0.0517,  0.0537,
                0.0557,  0.0577,  0.0596,  0.0616,  0.0636,  0.0656,  0.0676,
                ......
                0.4453,  0.4473,  0.4493,  0.4513,  0.4533,  0.4553,  0.4573,
                0.4592,  0.4612,  0.4632,  0.4652,  0.4672,  0.4692,  0.4712,
                0.4732,  0.4751,  0.4771,  0.4791,  0.4811,  0.4831,  0.4851,
                0.4871,  0.4891,  0.4911,  0.493 ,  0.495 ,  0.497 ,  0.499 ,
                ......
                0.7097,  0.7117,  0.7137,  0.7157,  0.7177,  0.7197,  0.7217,
                0.7237,  0.7256,  0.7276,  0.7296,  0.7316,  0.7336,  0.7356,
                0.7376,  0.7396,  0.7416,  0.7435,  0.7455,  0.7475,  0.7495,
                0.7515,  0.7535,  0.7555,  0.7575,  0.7594,  0.7614,  0.7634,
                ......
                0.9463,  0.9483,  0.9503,  0.9523,  0.9543,  0.9563,  0.9583,
                0.9602,  0.9622,  0.9642,  0.9662,  0.9682,  0.9702,  0.9722,
                0.9742,  0.9761,  0.9781,  0.9801,  0.9821,  0.9841,  0.9861,
                0.9881,  0.9901,  0.992 ,  0.994 ,  0.996 ,  0.998 ,  1.    ])
        """

        # 计算时间跳空能量分两部分组成：
        #   sw[0]：控制非时间加权的跳空能量所最终占比
        #   sw[1]：控制时间加权的跳空能量所最终占比
        w_power = (jump_tuple.power * sw[0]) + (weights[weight_index] *
                                                jump_tuple.power * sw[1])
        if w_power > power_threshold:
            # 最终加权能量大于阀值能量的进行能量替换，加入到ws_jump_lines中
            # noinspection PyProtectedMember
            jump_tuple = jump_tuple._replace(power=w_power)
            ws_jump_lines.append(jump_tuple)
    return ws_jump_lines


# -*- coding: utf-8 -*-

from ultron.sentry.Math.Accumulators.IAccumulators cimport Accumulator
from ultron.sentry.Analysis.SeriesValues cimport SeriesValues


cdef class SecurityValueHolder(object):

    cdef public list _dependency
    cdef public int _window
    cdef public Accumulator _holderTemplate
    cdef public int updated
    cdef public dict _innerHolders
    cdef public SeriesValues cached

    cpdef value_all(self)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef double value_by_name(self, name)
    cpdef shift(self, int n)
    cpdef transform(self, data, str name=*, str category_field=*, bint dropna=*)


cdef class SecuritySingleValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _compHolder
    cpdef push(self, dict data)


cdef class SecurityBinaryValueHolder(SecurityValueHolder):
    cdef public SecurityValueHolder _compHolder1
    cdef public SecurityValueHolder _compHolder2
    cpdef push(self, dict data)


cdef class SecurityStatelessSingleValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _compHolder

    cpdef push(self, dict data)


cdef class FilteredSecurityValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _filter
    cdef public SecurityValueHolder _computer

    cpdef double value_by_name(self, name)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef push(self, dict data)


cdef class IdentitySecurityValueHolder(SecurityValueHolder):

    cdef public object _value
    cdef set _symbols

    cpdef push(self, dict data)
    cpdef double value_by_name(self, name)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)


cdef class SecurityConstArrayValueHolder(SecurityValueHolder):
    cdef SeriesValues _values

    cpdef push(self, dict data)
    cpdef double value_by_name(self, name)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)


cdef class SecurityUnitoryValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _right
    cdef public object _op

    cpdef double value_by_name(self, name)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef push(self, dict data)


cdef class SecurityNegValueHolder(SecurityUnitoryValueHolder):
    pass


cdef class SecurityInvertValueHolder(SecurityUnitoryValueHolder):
    pass


cdef class SecurityCurrentValueHolder(SecurityValueHolder):
    cdef dict _symbol_values

    cpdef push(self, dict data)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef double value_by_name(self, name)


cdef class SecurityLatestValueHolder(SecurityValueHolder):
    cdef dict _symbol_values

    cpdef push(self, dict data)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef double value_by_name(self, name)


cpdef SecurityValueHolder build_holder(name)


cdef class SecurityCombinedValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _left
    cdef public SecurityValueHolder _right
    cdef public object _op

    cpdef double value_by_name(self, name)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef push(self, dict data)

cdef class SecurityXorValueHolder(SecurityCombinedValueHolder):

    cpdef double value_by_name(self, name)


cdef class SecurityAddedValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecuritySubbedValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityMultipliedValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityDividedValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityLtOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityLeOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityGtOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityGeOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityEqOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityNeOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityAndOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityOrOperatorValueHolder(SecurityCombinedValueHolder):
    pass


cdef class SecurityShiftedValueHolder(SecuritySingleValueHolder):
    pass

cdef class SecurityDeltaValueHolder(SecuritySingleValueHolder):
    pass


cdef class SecurityIIFValueHolder(SecurityValueHolder):

    cdef public SecurityValueHolder _flag
    cdef public SecurityValueHolder _left
    cdef public SecurityValueHolder _right

    cpdef double value_by_name(self, name)
    cpdef SeriesValues value_all(self)
    cpdef SeriesValues value_by_names(self, list names)
    cpdef push(self, dict data)


# -*- encoding:utf-8 -*-
"""
    技术指标工具基础模块
"""

import pandas as pd
from enum import Enum

from ultron.kdutils import date as date_util


class ECalcType(Enum):
    """
        技术指标技术方式类
    """
    """使用talib透传技术指标计算"""
    E_FROM_TA = 0
    """使用pandas等库实现技术指标计算"""
    E_FROM_PD = 1


"""彻底不用talib，完全都使用自己计算的指标结果"""
g_calc_type = ECalcType.E_FROM_PD



# -*- encoding:utf-8 -*-
"""
    买入择时示例因子：突破买入择时因子
"""
from ultron.ump.factor.buy.base import FactorBuyBase, FactorBuyXD, BuyCallMixin, BuyPutMixin


# noinspection PyAttributeOutsideInit
class FactorBuyBreak(FactorBuyBase, BuyCallMixin):
    """示例正向突破买入择时类，混入BuyCallMixin，即向上突破触发买入event"""

    def _init_self(self, **kwargs):
        """kwargs中必须包含: 突破参数xd 比如20，30，40天...突破"""
        # 突破参数 xd， 比如20，30，40天...突破, 不要使用kwargs.pop('xd', 20), 明确需要参数xq
        self.xd = kwargs['xd']
        # 在输出生成的orders_pd中显示的名字
        self.factor_name = '{}:{}'.format(self.__class__.__name__, self.xd)

    def fit_day(self, today):
        """
        针对每一个交易日拟合买入交易策略，寻找向上突破买入机会
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 忽略不符合买入的天（统计周期内前xd天）
        if self.today_ind < self.xd - 1:
            return None

        # 今天的收盘价格达到xd天内最高价格则符合买入条件
        if today.close == self.kl_pd.close[self.today_ind - self.xd +
                                           1:self.today_ind + 1].max():
            # 把突破新高参数赋值skip_days，这里也可以考虑make_buy_order确定是否买单成立，但是如果停盘太长时间等也不好
            self.skip_days = self.xd
            # 生成买入订单, 由于使用了今天的收盘价格做为策略信号判断，所以信号发出后，只能明天买
            return self.buy_tomorrow()
        return None


# noinspection PyAttributeOutsideInit
class FactorBuyXDBK(FactorBuyXD, BuyCallMixin):
    """示例继承FactorBuyXD完成正向突破买入择时类"""

    def fit_day(self, today):
        """
        针对每一个交易日拟合买入交易策略，寻找向上突破买入机会
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 今天的收盘价格达到xd天内最高价格则符合买入条件
        if today.close == self.xd_kl.close.max():
            return self.buy_tomorrow()
        return None


# noinspection PyAttributeOutsideInit
class FactorBuyPutBreak(FactorBuyBase, BuyPutMixin):
    """示例反向突破买入择时类，混入BuyPutMixin，即向下突破触发买入event，详情请查阅期货回测示例demo"""

    def _init_self(self, **kwargs):
        """kwargs中必须包含: 突破参数xd 比如20，30，40天...突破"""

        # 突破参数 xd， 比如20，30，40天...突破, 不要使用kwargs.pop('xd', 20), 明确需要参数xq
        self.xd = kwargs['xd']
        self.factor_name = '{}:{}'.format(self.__class__.__name__, self.xd)

    def fit_day(self, today):
        """
        针对每一个交易日拟合买入交易策略，寻找向下突破买入机会
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 忽略不符合买入的天（统计周期内前xd天）
        if self.today_ind < self.xd - 1:
            return None
        """
            与FactorBuyBreak区别就是买向下突破的，即min()
        """
        if today.close == self.kl_pd.close[self.today_ind - self.xd +
                                           1:self.today_ind + 1].min():
            self.skip_days = self.xd
            return self.buy_tomorrow()
        return None


# noinspection PyAttributeOutsideInit
class FactorBuyPutXDBK(FactorBuyXD, BuyPutMixin):
    """示例继承AbuFactorBuyXD完成反向突破买入择时类"""

    def fit_day(self, today):
        """
        针对每一个交易日拟合买入交易策略，寻找向上突破买入机会
        :param today: 当前驱动的交易日金融时间序列数据
        :return:
        """
        # 与AbuFactorBuyBreak区别就是买向下突破的，即min()
        if today.close == self.xd_kl.close.min():
            return self.buy_tomorrow()
        return None



# -*- coding: utf-8 -*-

from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union
import numpy as np
from .optimizers import (QuadraticOptimizer, TargetVolOptimizer)
from ultron.utilities.exceptions import PortfolioBuilderException


def _create_bounds(lbound, ubound, bm, risk_exposure, risk_target):
    if lbound is not None:
        lbound = lbound - bm
    if ubound is not None:
        ubound = ubound - bm

    if risk_exposure is not None:
        cons_mat = risk_exposure.T
        bm_risk = cons_mat @ bm

        clbound = (risk_target[0] - bm_risk).reshape((-1, 1))
        cubound = (risk_target[1] - bm_risk).reshape((-1, 1))
    else:
        cons_mat = None
        clbound = None
        cubound = None

    return lbound, ubound, cons_mat, clbound, cubound


def _create_result(optimizer, bm):
    if optimizer.status() == "optimal" or optimizer.status(
    ) == "optimal_inaccurate":
        return optimizer.status(), optimizer.feval(), optimizer.x_value() + bm
    else:
        raise (optimizer.status())


def mean_variance_builder(
        er: np.ndarray,
        risk_model: Dict[str, Union[None, np.ndarray]],
        bm: np.ndarray,
        lbound: Union[np.ndarray, float, None],
        ubound: Union[np.ndarray, float, None],
        risk_exposure: Optional[np.ndarray],
        risk_target: Optional[Tuple[np.ndarray, np.ndarray]],
        lam: float = 1.,
        linear_solver: str = 'deprecated') -> Tuple[str, float, np.ndarray]:
    lbound, ubound, cons_mat, clbound, cubound = _create_bounds(
        lbound, ubound, bm, risk_exposure, risk_target)
    if cons_mat is not None:
        cons_matrix = np.concatenate([cons_mat, clbound, cubound], axis=1)
    else:
        cons_matrix = None

    cov = risk_model['cov']
    special_risk = risk_model['idsync']
    risk_cov = risk_model['factor_cov']
    risk_exposure = risk_model['factor_loading']

    prob = QuadraticOptimizer(objective=-er,
                              cons_matrix=cons_matrix,
                              lbound=lbound,
                              ubound=ubound,
                              penalty=lam,
                              cov=cov,
                              factor_cov=risk_cov,
                              factor_load=risk_exposure,
                              factor_special=special_risk)

    if prob.status() == "optimal" or prob.status() == 'optimal_inaccurate':
        return prob.status(), prob.feval(), prob.x_value() + bm
    else:
        PortfolioBuilderException(prob.status())


def target_vol_builder(
        er: np.ndarray,
        risk_model: Dict[str, Union[None, np.ndarray]],
        bm: np.ndarray,
        lbound: Union[np.ndarray, float],
        ubound: Union[np.ndarray, float],
        risk_exposure: Optional[np.ndarray],
        risk_target: Optional[Tuple[np.ndarray, np.ndarray]],
        vol_target: float = 1.,
        linear_solver: str = 'ma27') -> Tuple[str, float, np.ndarray]:
    lbound, ubound, cons_mat, clbound, cubound = _create_bounds(
        lbound, ubound, bm, risk_exposure, risk_target)

    if cons_mat is not None:
        cons_matrix = np.concatenate([cons_mat, clbound, cubound], axis=1)
    else:
        cons_matrix = None

    cov = risk_model['cov']
    special_risk = risk_model['idsync']
    risk_cov = risk_model['factor_cov']
    risk_exposure = risk_model['factor_loading']

    prob = TargetVolOptimizer(objective=-er,
                              cons_matrix=cons_matrix,
                              lbound=lbound,
                              ubound=ubound,
                              target_vol=vol_target,
                              factor_cov=risk_cov,
                              factor_load=risk_exposure,
                              factor_special=special_risk,
                              cov=cov)
    if prob.status() == "optimal" or prob.status() == 'optimal_inaccurate':
        return prob.status(), prob.feval(), prob.x_value() + bm
    else:
        PortfolioBuilderException(prob.status())



# -*- coding: utf-8 -*-
import datetime
import six
from ultron.utilities.singleton import Singleton
from ultron.utilities.mlog import MLog

@six.add_metaclass(Singleton)
class ModelsSingleton(object):
    def __init__(self):
        __str__ = 'ModelsSingleton'
        self._cluster_map = {}
        
    def add_cluster(self, wid, cluster):
        self._cluster_map[wid] = cluster
    
    def verify_auth_token(self, wid, token):
        _token = None
        if wid in self._cluster_map:
            _token = self._cluster_map[wid]['token']
        return True if (token == _token and _token is not None and token is not None) else False
    
    def check_alive(self, wid):
        now_time = datetime.datetime.now()
        if wid in self._cluster_map:
            last_time = self._cluster_map[wid]['updatime']
        return True if ((now_time - last_time).seconds > 100) else False
    
    def set_alive(self, wid):
        now_time = datetime.datetime.now()
        if wid in self._cluster_map:
            self._cluster_map[wid]['updatime'] = now_time
            MLog.write().info('wid:%s, now_time:%s' % (str(wid), now_time.strftime('%Y-%m-%d %H:%M:%S')))
            
    
    def get_queue(self, wid):
        return self._cluster_map[wid]['queue_name'] if (wid in self._cluster_map) else None

    def get_all_queue(self):
        queue_list = []
        for wid, cluster in self._cluster_map.items():
            queue_list.append(cluster['queue_name'])
        return queue_list
            
        